from pystreed.data import PPGData
from pystreed.base import BaseSTreeDSolver
from sklearn.utils._param_validation import Interval, StrOptions
from typing import Optional
from pystreed.utils import _color_brew
import numpy as np

class STreeDPrescriptivePolicyGenerator(BaseSTreeDSolver):

    _parameter_constraints: dict = {**BaseSTreeDSolver._parameter_constraints, 
        "teacher_method": [StrOptions({"DR", "DM", "IPW"})]
    }

    def __init__(self, 
                 max_depth : int = 3,
                 max_num_nodes : Optional[int] = None,
                 min_leaf_node_size : int = 1,
                 teacher_method : str = "DR",
                 time_limit : float = 600,
                 feature_ordering : str = "gini", 
                 hyper_tune: bool = False,
                 use_branch_caching: bool = False,
                 use_dataset_caching: bool = True,
                 use_terminal_solver: bool = True,
                 use_similarity_lower_bound: bool = True,
                 use_upper_bound: bool = True,
                 use_lower_bound: bool = True,
                 upper_bound: float = 2**31 - 1,
                 verbose : bool = False,
                 random_seed : int = 27,
                 continuous_binarize_strategy: str = 'quantile',
                 n_thresholds: int = 5,
                 n_categories: int = 5):
        """
        Construct a STreeDPrescriptivePolicyGenerator

        Parameters:
            max_depth: the maximum depth of the tree
            max_num_nodes: the maximum number of branching nodes of the tree
            min_leaf_node_size: the minimum number of training instance that should end up in every leaf node
            teacher_method: the teacher objective function. Either direct method ("DM"), inverse propensity weighting ("IPW") or doubly robust ("DR")
            time_limit: the time limit in seconds for fitting the tree
            feature_ordering: heuristic for the order that features are checked. Default: "gini", alternative: "in-order": the order in the given data
            hyper_tune: Use five-fold validation to tune the size of the tree to prevent overfitting
            use_branch_caching: Enable/Disable branch caching (typically the slower caching strategy. May be faster in some scenario's)
            use_dataset_caching: Enable/Disable dataset caching (typically the faster caching strategy)
            use_terminal_solver: Enable/Disable the depth-two solver (Enabled typically results in a large runtime advantage)
            use_similarity_lower_bound: Enable/Disable the similarity lower bound (Enabled typically results in a large runtime advantage)
            use_upper_bound: Enable/Disable the use of upper bounds (Enabled is typically faster)
            use_lower_bound: Enable/Disable the use of lower bounds (Enabled is typically faster)
            upper_bound: Search for a tree better than the provided upper bound
            verbose: Enable/Disable verbose output
            random_seed: the random seed used by the solver (for example when creating folds)
            continuous_binarization_strategy: the strategy used for binarizing continuous features
            n_thresholds: the number of thresholds to use per continuous feature
            n_categories: the number of categories to use per categorical feature
        """
        BaseSTreeDSolver.__init__(self, "prescriptive-policy", 
            max_depth=max_depth,
            max_num_nodes=max_num_nodes,
            min_leaf_node_size=min_leaf_node_size,
            time_limit = time_limit,
            cost_complexity=0,
            feature_ordering=feature_ordering,
            hyper_tune = hyper_tune,
            use_branch_caching=use_branch_caching,
            use_dataset_caching=use_dataset_caching,
            use_terminal_solver=use_terminal_solver,
            use_similarity_lower_bound=use_similarity_lower_bound,
            use_upper_bound=use_upper_bound,
            use_lower_bound=use_lower_bound,
            upper_bound=upper_bound,
            verbose=verbose,
            random_seed=random_seed,
            continuous_binarize_strategy=continuous_binarize_strategy,
            n_thresholds=n_thresholds,
            n_categories=n_categories)
        self.teacher_method = teacher_method
        self.n_classes_ = None

    def _initialize_param_handler(self):
        super()._initialize_param_handler()
        self._params.ppg_teacher_method = self.teacher_method
    
    def _check_data(self, y_in, reset=True):
        y_out = []
        if isinstance(y_in, np.ndarray):
            extra_data = []
            if len(y_in.shape) == 2 and y_in.shape[1] >= 5:
                if reset or self.n_classes_ is None:
                    n_classes = len(np.unique(y_in[:, 0]))
                else:
                    n_classes = self.n_classes_
                for row in y_in:
                    historic_treatment = int(row[0])
                    historic_outcome = float(row[1])
                    propensity_score = float(row[2])
                    if y_in.shape[1] == 4 + n_classes * 2:
                        regress_compare_yhat = [float(row[3 + i]) for i in range(n_classes)]
                        optimal_treatment = int(row[3+n_classes])
                        counterfactual_y = [float(row[4 + n_classes + i]) for i in range(n_classes)]
                        
                        extra_data.append(PPGData(historic_treatment, historic_outcome, 
                                                propensity_score, regress_compare_yhat,
                                                optimal_treatment, counterfactual_y))
                    elif y_in.shape[1] == 3 + n_classes:
                        regress_compare_yhat = [float(row[3 + i]) for i in range(n_classes)]
                        extra_data.append(PPGData(historic_treatment, historic_outcome, 
                                                propensity_score, regress_compare_yhat))
                    else:
                        raise ValueError(f"The array does not have the right shape. Expected shape is either {3+n_classes} columns or {4 + 2*n_classes} columns.")
            else:
                raise ValueError(f"The array does not have the right shape. Expected shape is a two-dimensional array with at least 5 columns.")
        else:
            extra_data = y_in
        
        for i, inst in enumerate(extra_data):
            if not isinstance(inst, PPGData):
                raise ValueError("Each instance in y is expected to be of class PPGData.")
            y_out.append(inst.historic_treatment)

        if reset or self.n_classes_ is None:
            self.n_classes_ = len(np.unique(y_out))
        return np.array(y_out, dtype=np.intc), extra_data

    def _process_extra_data(self, X, extra_data):
        return extra_data

    def fit(self, X, y, categorical=None):
        """
        Fits a STreeD model to the given training data.

        Args:
            x : array-like, shape = (n_samples, n_features)
            Data matrix

            y : array-like with shape (n_samples, n_ppg_columns) 
            Target vector. n_ppg_columns depends on the number of possible treatments K
            * Column 0 is the historic label k (treatment)
            * Column 1 is the historic outcome y 
            * Column 2 is the propensity score mu
            * Column 3 .. 3+K-1 is the regress and compare y_hat prediction
            The rest of the data is optional (for testing)
            * Column 3 + K is the optimal treatment
            * Column 4 + K .. 4 + 2K - 1 is the counterfactual outcome y

            categorical : array-like, 
            List of column names that are categorical

        Returns:
            STreeDPrescriptivePolicyGenerator

        Raises:
            ValueError: If x or y is None or if they have different number of rows.
        """
        y, extra_data = self._check_data(y)
        return super().fit(X, y, extra_data, categorical)
    
    def predict(self, X, extra_data=None):
        """
        Predicts the target variable for the given input feature data.

        Args:
            x : array-like, shape = (n_samples, n_features)
            Data matrix
            extra_data (optional) : array-like with shape (n_samples, n_ppg_columns) 
            Target vector. n_ppg_columns depends on the number of possible treatments K
            * Column 0 is the historic label k (treatment)
            * Column 1 is the historic outcome y 
            * Column 2 is the propensity score mu
            * Column 3 .. 3+K-1 is the regress and compare y_hat prediction
            The rest of the data is optional (for testing)
            * Column 3 + K is the optimal treatment
            * Column 4 + K .. 4 + 2K - 1 is the counterfactual outcome y
            
        Returns:
            numpy.ndarray: A 1D array that represents the predicted target variable of the test data.
                The i-th element in this array corresponds to the predicted target variable for the i-th instance in `x`.
        """
        if extra_data is None:
            extra_data = [PPGData() for i in range(len(X))]
        _, extra_data = self._check_data(extra_data, reset=False)
        return super().predict(X, extra_data)
    
    def score(self, X, y_test):
        """
        Computes the score for the given input feature data

        Args:
            x : array-like, shape = (n_samples, n_features)
            Data matrix
            y_test : array-like with shape (n_samples, n_ppg_columns) 
            Target vector. n_ppg_columns depends on the number of possible treatments K
            * Column 0 is the historic label k (treatment)
            * Column 1 is the historic outcome y 
            * Column 2 is the propensity score mu
            * Column 3 .. 3+K-1 is the regress and compare y_hat prediction
            * Column 3 + K is the optimal treatment
            * Column 4 + K .. 4 + 2K - 1 is the counterfactual outcome y

        Returns:
            The score
        """
        y_test, extra_data = self._check_data(y_test, reset=False)
        return super().score(X, y_test, extra_data)

    def _export_dot_leaf_node(self, fh, node, node_id, label_names, train_data):
        if not hasattr(self, "_colors"):
            self._colors = _color_brew(self.n_classes_)
        color = self._colors[node.label]
        return super()._export_dot_leaf_node(fh, node, node_id, label_names, train_data, color=color)


import pandas as pd
import numpy as np
from pandas.api.types import is_numeric_dtype
from sklearn.utils._param_validation import Interval, StrOptions
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import _check_feature_names_in, check_array, check_is_fitted
from sklearn.tree import DecisionTreeRegressor, DecisionTreeClassifier
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.model_selection import GridSearchCV
import sklearn
import numbers
import itertools
import math
import warnings
import collections

def get_column_types(X, categorical_columns=None):
    if categorical_columns is None: 
        categorical_columns = []
    binary_columns = []
    if isinstance(X, pd.DataFrame):
        categorical_columns = categorical_columns + [col for col in X.columns if not col in categorical_columns and not is_numeric_dtype(X[col])]
        binary_columns = [col for col in X.columns if not set(X[col].unique()) - set([0,1])] 
        continuous_columns = [col for col in X.columns if not col in categorical_columns and not col in binary_columns]
    elif isinstance(X, np.ndarray):
        binary_columns = [col for col in range(X.shape[1]) if not set(np.unique(X[:, col])) - set([0,1])]
        continuous_columns = [col for col in range(X.shape[1]) if not col in categorical_columns and not col in binary_columns]
    else:
        raise ValueError("X should be either a numpy array or a pandas data frame.")
    return binary_columns, categorical_columns, continuous_columns

def _escape(obj):
    if isinstance(obj, list):
        return [_escape(j) for j in obj]
    elif isinstance(obj, str):
        return obj.replace("'", "\\'")
    else:
        return obj

class Binarizer:

    def __init__(self, continuous_strategy, n_thresholds, n_categories, categorical_columns, max_num_binary_features = None):
        """
        Initialize a binarizer

        Parameters:
        continuous_strategy: str, the binarization strategy to use for continuous features. Either "quantile", "uniform" or "tree"
        n_thresholds: int or array-like, the number of tresholds to use for continuous features. If an array, it should have a number for every continuous feature
        n_categories: int or array-like, the number of categories to turn every categorical feature into. If an array, it should have a number for every categorical feature
        categorical_columns: array-like, the names of the categorical feature columns
        """
        self.continuous_strategy = continuous_strategy
        self.n_thresholds = n_thresholds
        self.n_categories = n_categories
        self.max_num_binary_features = max_num_binary_features
        self.categorical_columns = [] if categorical_columns is None else categorical_columns
        self.binary_columns = []
        self.continuous_columns = []
        self.categorical_binarizer = None
        self.continuous_binarizer = None
        self.binary_feature_limiter = None

    def fit(self, X, y):
        X_cat = None
        X_bin = None
        X_cont = None
        self.binary_columns, self.categorical_columns, self.continuous_columns = get_column_types(X, self.categorical_columns)
        if isinstance(X, pd.DataFrame):
            if len(self.categorical_columns) > 0:
                X_cat = X[self.categorical_columns]
            if len(self.binary_columns) > 0:
                X_bin = X[self.binary_columns]
            if len(self.continuous_columns) > 0:
                X_cont = X[self.continuous_columns]
        else:
            if len(self.categorical_columns) > 0:
                X_cat = X[:, self.categorical_columns]
            if len(self.binary_columns) > 0:
                X_bin = X[:, self.binary_columns]
            if len(self.continuous_columns) > 0:
                X_cont = X[:, self.continuous_columns]

        if len(self.categorical_columns) > 0:
            self.categorical_binarizer = CategoricalBinarizer(self.n_categories) 
            self.categorical_binarizer.fit(X_cat)
        if len(self.continuous_columns) > 0:
            self.continuous_binarizer = KThresholdBinarizer(self.continuous_strategy, self.n_thresholds)
            self.continuous_binarizer.fit(X_cont, y)

        if not self.max_num_binary_features is None:
            if y is None:
                warnings.warn("Cannot limit the number of binary features, because the class label y is None")
                return
            X_trans = self.transform(X)
            self.binary_feature_limiter = RandomForestFeatureSelector(self.max_num_binary_features)
            self.binary_feature_limiter.fit(X_trans, y)

    def transform(self, X):
        parts = []

        if isinstance(X, pd.DataFrame):
            if len(self.categorical_columns) > 0:
                X_cat = X[self.categorical_columns]
            if len(self.binary_columns) > 0:
                X_bin = X[self.binary_columns]
            if len(self.continuous_columns) > 0:
                X_cont = X[self.continuous_columns]
        else:
            if len(self.categorical_columns) > 0:
                X_cat = X[:, self.categorical_columns]
            if len(self.binary_columns) > 0:
                X_bin = X[:, self.binary_columns]
            if len(self.continuous_columns) > 0:
                X_cont = X[:, self.continuous_columns]
        if len(self.binary_columns) > 0:
            if isinstance(X_bin, pd.DataFrame) and not is_numeric_dtype(X_bin):
                X_bin.columns = [f"Binary Feature {i+1}" if not isinstance(X_bin.columns[i], str) else X_bin.columns[i] for i in range(X_bin.shape[1])]
            parts.append(X_bin)
        if len(self.categorical_columns) > 0:
            X_cat = self.categorical_binarizer.transform(X_cat)
            parts.append(X_cat)
        if len(self.continuous_columns) > 0:
            X_cont = self.continuous_binarizer.transform(X_cont)
            parts.append(X_cont)
                
        if isinstance(X, pd.DataFrame):
            for p in parts:
                p.index = X.index
            X_trans = pd.concat(parts, axis=1)
        else:
            X_trans = np.concatenate(parts, axis=1)
        if not self.binary_feature_limiter is None:
            return self.binary_feature_limiter.transform(X_trans)
        return X_trans

class CategoricalBinarizer(BaseEstimator, TransformerMixin):

    _parameter_constraints: dict = {
        "n_categories": [Interval(numbers.Integral, 1, None, closed="left"), "array-like"],
        "minimum_coverage": [Interval(numbers.Real, 0, 1, closed="both")]
    }
    
    def __init__(self, n_categories=10, minimum_coverage=0.01):
        self.n_categories = n_categories
        self.minimum_coverage = minimum_coverage

    def _validate_n_categories(self, n_features):
        """Returns n_categories_, the number of categories per feature."""
        orig_categories = self.n_categories
        if isinstance(orig_categories, numbers.Integral):
            return np.full(n_features, orig_categories, dtype=int)

        n_categories = check_array(orig_categories, dtype=int, copy=True, ensure_2d=False)

        if n_categories.ndim > 1 or n_categories.shape[0] != n_features:
            raise ValueError("n_categories must be a scalar or array of shape (n_features,).")

        bad_n_categories_value = (n_categories < 1) | (n_categories != orig_categories)

        violating_indices = np.where(bad_n_categories_value)[0]
        if violating_indices.shape[0] > 0:
            indices = ", ".join(str(i) for i in violating_indices)
            raise ValueError(
                "{} received an invalid number "
                "of categories at indices {}. Number of categories "
                "must be at least 1, and must be an int.".format(
                    CategoricalBinarizer.__name__, indices
                )
            )
        return n_categories
    
    def fit(self, X, y=None):
        self._validate_params()
        with warnings.catch_warnings():
            warnings.filterwarnings(action="ignore", category=FutureWarning)
            X = self._validate_data(X, dtype=None)

        n_samples, n_features = X.shape
        n_categories = self._validate_n_categories(n_features)
        categories = []
        column_names = []
        for jj in range(n_features):
            column = X[:, jj]
            if hasattr(self, "feature_names_in_") and len(self.feature_names_in_) == n_features:
                column_name = self.feature_names_in_[jj]
            else:
                column_name = f"Feature {jj + 1}"
            column_dtype = column.dtype
            counter = collections.Counter(column)
            
            if n_categories[jj] <= 1 or len(counter) <= 1:
                n_categories[jj] = 0
                categories.append([])
                column_names.append([])
                continue
            elif n_categories[jj] == 2 or len(counter) == 2:
                value = counter.most_common(2)[1][0] # get the key of the second most common value
                n_categories[jj] = 1
                categories.append([value])
                escaped_value = _escape(value)
                if isinstance(escaped_value, str):
                    column_names.append([f"{column_name} == '{value}'"])
                else:
                    column_names.append([f"{column_name} == {value}"])
                continue
            values = list(counter.keys())
            if len(values) > n_categories[jj]:
                values = values[:n_categories[jj] - 1] + [values[n_categories[jj] - 1:]]
            n_categories[jj] = len(values)
            categories.append(values)
            new_column_names = []
            escape_values = _escape(values)
            for v in escape_values:
                if isinstance(v, list):
                    new_column_names.append("{} in [{}]".format(column_name, \
                        ",".join([f"'{o}'" if isinstance(o, str) else str(o) for o in v])))
                elif isinstance(v, str):
                    new_column_names.append("{} == '{}'".format(column_name, v))
                else:
                    new_column_names.append("{} == {}".format(column_name, v))
            column_names.append(new_column_names)

        self.n_categories_ = n_categories
        self.categories_ = categories
        self.column_names_ = column_names

    def transform(self, X, y=None):
        check_is_fitted(self)
        with warnings.catch_warnings():
            warnings.filterwarnings(action="ignore", category=FutureWarning)
            X = self._validate_data(X, reset=False, dtype=None)

        n_samples, n_features = X.shape
        sum_of_bin_features = sum(self.n_categories_)
        bin_X = np.zeros((n_samples, sum_of_bin_features), dtype=np.intc)
        col_ix = 0
        for jj in range(n_features):
            if self.n_categories_[jj] == 0: continue
            column = X[:, jj]
            for c_ix in range(self.n_categories_[jj]):
                category_val = self.categories_[jj][c_ix]
                if isinstance(category_val, list):
                    bin_X[:, col_ix] = np.array([v in category_val for v in column])
                else:
                    bin_X[:, col_ix] = column == category_val
                col_ix += 1
        column_names = itertools.chain.from_iterable(self.column_names_)
        return pd.DataFrame(bin_X, columns=column_names)

class KThresholdBinarizer(BaseEstimator, TransformerMixin):
    
    _parameter_constraints: dict = {
        "strategy": [StrOptions({"tree", "quantile", "uniform"})],
        "n_thresholds": [Interval(numbers.Integral, 1, None, closed="left"), "array-like"]
    }

    def __init__(self, strategy="quantile", n_thresholds=10):
        self.strategy = strategy
        self.n_thresholds = n_thresholds

    def _validate_n_thresholds(self, n_features):
        """Returns n_thresholds_, the number of thresholds per feature."""
        orig_thresholds = self.n_thresholds
        if isinstance(orig_thresholds, numbers.Integral):
            return np.full(n_features, orig_thresholds, dtype=int)

        n_thresholds = check_array(orig_thresholds, dtype=int, copy=True, ensure_2d=False)

        if n_thresholds.ndim > 1 or n_thresholds.shape[0] != n_features:
            raise ValueError("n_thresholds must be a scalar or array of shape (n_features,).")

        bad_nthresholds_value = (n_thresholds < 1) | (n_thresholds != orig_thresholds)

        violating_indices = np.where(bad_nthresholds_value)[0]
        if violating_indices.shape[0] > 0:
            indices = ", ".join(str(i) for i in violating_indices)
            raise ValueError(
                "{} received an invalid number "
                "of thresholds at indices {}. Number of thresholds "
                "must be at least 1, and must be an int.".format(
                    KThresholdBinarizer.__name__, indices
                )
            )
        return n_thresholds

    def fit(self, X, y=None):
        with warnings.catch_warnings():
            warnings.filterwarnings(action="ignore", category=FutureWarning)
            X = self._validate_data(X, dtype="numeric")
            
        if self.strategy == 'tree':
            if y is None:
                raise ValueError("Binarizing with the tree strategy requires the labels y.")
            with warnings.catch_warnings():
                warnings.filterwarnings(action="ignore", category=FutureWarning)
                y = check_array(y, ensure_2d=False)
            if X.shape[0] != y.shape[0]:
                raise ValueError("Number of samples in X and y are not the same.")

        n_samples, n_features = X.shape
        n_thresholds = self._validate_n_thresholds(n_features)
        
        thresholds = []
        column_names = []
        for jj in range(n_features):
            column = X[:, jj]
            if hasattr(self, "feature_names_in_") and len(self.feature_names_in_) == n_features:
                column_name = self.feature_names_in_[jj]
            else:
                column_name = f"Feature {jj + 1}"
            column_dtype = column.dtype

            col_min, col_max = column.min(), column.max()
            if n_thresholds[jj] == 0 or col_min == col_max:
                n_thresholds[jj] = 0
                thresholds.append([])
                column_names.append([])
                continue
            if self.strategy in ["quantile", "uniform"]:
                if len(np.unique(column)) <=  n_thresholds[jj] + 1:
                    uniques = np.unique(column)
                    cutpoints = [(uniques[i] + uniques[i+1]) / 2 for i in range(len(uniques) - 1)]
                elif self.strategy == "uniform":
                    uniques = np.linspace(col_min, col_max, n_thresholds[jj] + 1)
                    cutpoints = np.unique([(uniques[i] + uniques[i+1]) / 2 for i in range(len(uniques) - 1)])
                else:
                    quantiles = np.linspace(0, 1, n_thresholds[jj] + 2)[1:-1]
                    cutpoints = np.unique(np.quantile(column, quantiles))
            elif self.strategy == 'tree':
                _X = np.array(column).reshape(-1, 1)
                is_classification = issubclass(y.dtype.type, np.integer) and len(np.unique(y)) <= 10
                Ml_model = DecisionTreeClassifier if is_classification else DecisionTreeRegressor
                #error_function = 'accuracy' if is_classification else "neg_mean_squared_error"
                model = Ml_model(max_leaf_nodes=n_thresholds[jj] + 1)
                model.fit(_X, y)
                cutpoints = sorted([t for t in model.tree_.threshold if t != sklearn.tree._tree.TREE_UNDEFINED])
            n_thresholds[jj] = len(cutpoints)
            thresholds.append(cutpoints)
            column_names.append([f"{column_name} <= {c}" for c in cutpoints])
        
        self.n_thresholds_ = n_thresholds
        self.thresholds_ = thresholds
        self.column_names_ = column_names
    
        return self

    def transform(self, X, y=None):
        check_is_fitted(self)
        with warnings.catch_warnings():
            warnings.filterwarnings(action="ignore", category=FutureWarning)
            X = self._validate_data(X, reset=False)

        n_samples, n_features = X.shape
        sum_of_bin_features = sum(self.n_thresholds_)
        bin_X = np.zeros((n_samples, sum_of_bin_features), dtype=np.intc)
        col_ix = 0
        for jj in range(n_features):
            if self.n_thresholds_[jj] == 0: continue
            column = X[:, jj]
            for t_ix in range(self.n_thresholds_[jj]):
                bin_X[:, col_ix] = column <= self.thresholds_[jj][t_ix]
                col_ix += 1
        column_names = itertools.chain.from_iterable(self.column_names_)
        return pd.DataFrame(bin_X, columns=column_names)
    
class RandomForestFeatureSelector(BaseEstimator, TransformerMixin):
    
    def __init__(self, max_num_binary_features : int):
        self.max_num_binary_features = max_num_binary_features

    def fit(self, X, y):
        is_classification = issubclass(y.dtype.type, np.integer) and len(np.unique(y)) <= 10
        rf_model = RandomForestClassifier if is_classification else RandomForestRegressor
        rf = rf_model(n_estimators=100)
        rf.fit(X, y)
        importances = rf.feature_importances_        
        indices = np.argsort(importances)[-self.max_num_binary_features:]

        if isinstance(X, pd.DataFrame):
            self.select_features_ = X.columns[indices]
        else:
            self.select_features_ = indices

    def transform(self, X):
        check_is_fitted(self, "select_features_")
        if isinstance(X, pd.DataFrame):
            return X[self.select_features_]
        else:
            return X[:, self.select_features_]



from pystreed.base import BaseSTreeDSolver
from pystreed.data import SAData
from sklearn.utils.validation import check_is_fitted
from sksurv.util import check_array_survival
from sksurv.tree._criterion import get_unique_times
from sksurv.tree.tree import _array_to_step_function
from typing import Optional
import numpy as np
import math
from pathlib import Path
import os


class STreeDSurvivalAnalysis(BaseSTreeDSolver):
    """
    An optimal survival tree learner using dynamic programming

    This class follows the design of the scikit-survival package.
    See https://scikit-survival.readthedocs.io/en/stable/user_guide/00-introduction.html for an introduction
    Comments and method definitions taken from https://github.com/sebp/scikit-survival/blob/v0.21.0/sksurv/tree/tree.py
    """

    def __init__(self,
                 max_depth : int = 3,
                 max_num_nodes : Optional[int] = None,
                 min_leaf_node_size : int = 1,
                 time_limit : float = 600,
                 hyper_tune : bool = False,
                 use_branch_caching: bool = False,
                 use_dataset_caching: bool = True,
                 use_terminal_solver: bool = True,
                 use_upper_bound: bool = True,
                 use_lower_bound: bool = True,
                 upper_bound: float = 2**31 - 1,
                 verbose : bool = False,
                 random_seed: int = 27, 
                 continuous_binarize_strategy: str = 'quantile',
                 n_thresholds: int = 5,
                 n_categories: int = 5):
        BaseSTreeDSolver.__init__(self, "survival-analysis",
            max_depth=max_depth,
            max_num_nodes=max_num_nodes,
            min_leaf_node_size=min_leaf_node_size,
            time_limit=time_limit,
            feature_ordering="in-order",
            hyper_tune = hyper_tune,
            use_branch_caching=use_branch_caching,
            use_dataset_caching=use_dataset_caching,
            use_terminal_solver=use_terminal_solver,
            use_similarity_lower_bound=False,
            use_upper_bound=use_upper_bound,
            use_lower_bound=use_lower_bound,
            upper_bound=upper_bound,
            verbose=verbose,
            random_seed=random_seed,
            continuous_binarize_strategy=continuous_binarize_strategy,
            n_thresholds=n_thresholds,
            n_categories=n_categories)
        self._label_type = np.double
    """
        Construct a STreeDSurvivalAnalysis solver

        Parameters:
            max_depth: the maximum depth of the tree
            max_num_nodes: the maximum number of branching nodes of the tree
            min_leaf_node_size: the minimum number of training instance that should end up in every leaf node
            time_limit: the time limit in seconds for fitting the tree
            cost_complexity: the cost of adding a branch node, expressed as a percentage. E.g., 0.01 means a branching node may be added if it increases the training accuracy by at least 1%.
            feature_ordering: heuristic for the order that features are checked. Default: "gini", alternative: "in-order": the order in the given data
            hyper_tune: Use five-fold validation to tune the size of the tree to prevent overfitting
            use_branch_caching: Enable/Disable branch caching (typically the slower caching strategy. May be faster in some scenario's)
            use_dataset_caching: Enable/Disable dataset caching (typically the faster caching strategy)
            use_terminal_solver: Enable/Disable the depth-two solver (Enabled typically results in a large runtime advantage)
            use_upper_bound: Enable/Disable the use of upper bounds (Enabled is typically faster)
            use_lower_bound: Enable/Disable the use of lower bounds (Enabled is typically faster)
            upper_bound: Search for a tree better than the provided upper bound
            verbose: Enable/Disable verbose output
            random_seed: the random seed used by the solver (for example when creating folds)
            continuous_binarization_strategy: the strategy used for binarizing continuous features
            n_thresholds: the number of thresholds to use per continuous feature
            n_categories: the number of categories to use per categorical feature
        """

    def _process_extra_data(self, X, events):
        """
        Preprocess the events into a list of SAData objects. Initialize the hazard value with -1
        """
        if events is None:
            return [SAData(1, -1.0) for x in X]    
        return [SAData(e, -1.0) for e in events]

    def fit(self, X, y, categorical=None):
        """
        Fits a STreeD survival tree model to the given training data.

         Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Data matrix

        y : structured array, shape = (n_samples,)
            A structured array containing the binary event indicator
            as first field, and time of event or time of censoring as
            second field.

        categorical : array-like, 
            List of column names that are categorical

        Returns
        -------
        self
        """
        if X.shape[0] != y.shape[0]:
            raise ValueError('x and y have different number of rows')
        events, times = check_array_survival(X, y)
        self._baseline_hazard_function = self._compute_baseline_cumulative_hazard_function(events, times)
        self.unique_times_, self.is_event_time_ = get_unique_times(times, events)
        self.n_outputs_ = self.unique_times_.shape[0]
        self.n_classes_ = np.ones(self.n_outputs_, dtype=np.intp) * 2

        return super().fit(X, times, events, categorical)
    
    def score(self, X, y_true):
        """
        Computes the score for the given input feature data and the true

        Args:
            x : array-like, shape = (n_samples, n_features)
            Data matrix
            
            y : structured array, shape = (n_samples,)
            A structured array containing the binary event indicator
            as first field, and time of event or time of censoring as
            second field.

        Returns:
            The score
        """
        check_is_fitted(self, "fit_result")
        if X.shape[0] != y_true.shape[0]:
            raise ValueError('x and y have different number of rows')
        events, times = check_array_survival(X, y_true)
        return super().score(X, times, events)

    def predict(self, X, extra_data=None):
        """Predict risk score.

        The risk score is the total number of events, which can
        be estimated by the sum of the estimated cumulative
        hazard function :math:`\\hat{H}_h` in terminal node :math:`h`.

        .. math::

            \\sum_{j=1}^{n(h)} \\hat{H}_h(T_{j} \\mid x) ,

        where :math:`n(h)` denotes the number of distinct event times
        of samples belonging to the same terminal node as :math:`x`.

        Parameters
        ----------
        X : array-like or sparse matrix, shape = (n_samples, n_features)
            Data matrix.

        check_input : boolean, default: True
            Allow to bypass several input checking.
            Don't use this parameter unless you know what you do.

        Returns
        -------
        risk_scores : ndarray, shape = (n_samples,)
            Predicted risk scores.
        """
        chf = self.predict_cumulative_hazard_function(X, return_array=True)
        return chf[:, self.is_event_time_].sum(1)

    def _compute_baseline_cumulative_hazard_function(self, events, times):
        """
        Compute an estimate for the baseline cumulative hazard function
        using the Nelson-Aalen estimator
        """
        ts = {}
        n = len(events)
        for d, t in zip(events, times):
            if t not in ts:
                ts[t] = [0, 0]
            ts[t][d] += 1

        d = {0: 1 / (n + 1)}
        at_risk = n
        sum = 0
        for t in sorted(ts.keys()):
            left, died = ts[t]
            sum += died / at_risk
            at_risk -= died + left

            if died > 0:
                d[t] = sum

        keys = [*d.keys()]

        def f(x):
            try:
                return d[x]
            except:
                pass

            a = 0
            b = len(keys) - 1
            while a != b:
                mid = (a + b + 1) // 2
                if keys[mid] > x:
                    b = mid - 1
                else:
                    a = mid
            return d[keys[a]]

        return f

    def predict_cumulative_hazard_function(self, X, extra_data=None, return_array=False):
        """Predict cumulative hazard function.

        The cumulative hazard function (CHF) for an individual
        with feature vector :math:`x` is computed from
        all samples of the training data that are in the
        same terminal node as :math:`x`.
        It is estimated by the Nelson–Aalen estimator.

        Parameters
        ----------
        X : array-like or sparse matrix, shape = (n_samples, n_features)
            Data matrix.

        check_input : boolean, default: True
            Allow to bypass several input checking.
            Don't use this parameter unless you know what you do.

        return_array : boolean, default: False
            If set, return an array with the cumulative hazard rate
            for each `self.unique_times_`, otherwise an array of
            :class:`sksurv.functions.StepFunction`.

        Returns
        -------
        cum_hazard : ndarray
            If `return_array` is set, an array with the cumulative hazard rate
            for each `self.unique_times_`, otherwise an array of length `n_samples`
            of :class:`sksurv.functions.StepFunction` instances will be returned.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> from sksurv.datasets import load_whas500
        >>> from sksurv.tree import SurvivalTree

        Load and prepare the data.

        >>> X, y = load_whas500()
        >>> X = X.astype(float)

        Fit the model.

        >>> estimator = SurvivalTree().fit(X, y)

        Estimate the cumulative hazard function for the first 5 samples.

        >>> chf_funcs = estimator.predict_cumulative_hazard_function(X.iloc[:5])

        Plot the estimated cumulative hazard functions.

        >>> for fn in chf_funcs:
        ...    plt.step(fn.x, fn(fn.x), where="post")
        ...
        >>> plt.ylim(0, 1)
        >>> plt.show()
        """
        thetas = super().predict(X, extra_data)
        arr = np.array([
                [theta * self._baseline_hazard_function(t) for t in self.unique_times_]
                 for theta in thetas])
        if return_array: return arr
        return _array_to_step_function(self.unique_times_, arr)

    def predict_survival_function(self, X, extra_data=None, return_array=False):
        """Predict survival function.

        The survival function for an individual
        with feature vector :math:`x` is computed from
        all samples of the training data that are in the
        same terminal node as :math:`x`.
        It is estimated by the Kaplan-Meier estimator.

        Parameters
        ----------
        X : array-like or sparse matrix, shape = (n_samples, n_features)
            Data matrix.

        check_input : boolean, default: True
            Allow to bypass several input checking.
            Don't use this parameter unless you know what you do.

        return_array : boolean, default: False
            If set, return an array with the probability
            of survival for each `self.unique_times_`,
            otherwise an array of :class:`sksurv.functions.StepFunction`.

        Returns
        -------
        survival : ndarray
            If `return_array` is set, an array with the probability of
            survival for each `self.unique_times_`, otherwise an array of
            length `n_samples` of :class:`sksurv.functions.StepFunction`
            instances will be returned.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> from sksurv.datasets import load_whas500
        >>> from sksurv.tree import SurvivalTree

        Load and prepare the data.

        >>> X, y = load_whas500()
        >>> X = X.astype(float)

        Fit the model.

        >>> estimator = SurvivalTree().fit(X, y)

        Estimate the survival function for the first 5 samples.

        >>> surv_funcs = estimator.predict_survival_function(X.iloc[:5])

        Plot the estimated survival functions.

        >>> for fn in surv_funcs:
        ...    plt.step(fn.x, fn(fn.x), where="post")
        ...
        >>> plt.ylim(0, 1)
        >>> plt.show()
        """
        thetas = super().predict(X, extra_data)
        arr = np.array([
                [math.exp(-theta * self._baseline_hazard_function(t)) for t in self.unique_times_]
                 for theta in thetas])
        if return_array: return arr
        return _array_to_step_function(self.unique_times_, arr)
    

    def _export_dot_leaf_node(self, fh, node, node_id, label_names, train_data):
        try:
            import matplotlib.pyplot as plt
            theta = node.label
            arr = np.array([[math.exp(-theta * self._baseline_hazard_function(t)) for t in self.unique_times_]])
            fn = _array_to_step_function(self.unique_times_, arr)[0]
            plt.figure(figsize=(2,2))
            plt.step(fn.x, fn(fn.x), where="post")
            plt.ylim(0, 1)
            plt.ylabel("Probability of survival $\hat{S}(t)$")
            plt.xlabel("time $t$")
            filename = self.__export_directory / f"leaf_KM_{node_id}.png"
            if not os.path.exists(self.__export_directory):
                os.makedirs(self.__export_directory)
            plt.savefig(filename, dpi=300, bbox_inches="tight")
            fh.write(f"{node_id}  [label=\"\", image=\"{filename}\", width=2, height=2, fixedsize=true] ;\n")
        except:
            return super()._export_dot_leaf_node(fh, node, node_id, label_names, train_data)


    def export_dot(self, filename, feature_names=None, label_names=None):
        """
        Export the tree as a .dot file (for plotting)
        """
        self.__export_directory = Path(filename).parent.absolute() / "leaf-distributions"
        return super().export_dot(filename, feature_names, label_names)
        



from sklearn.utils._param_validation import Interval, StrOptions
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_array, check_is_fitted
from .cstreed import initialize_streed_solver, ParameterHandler, SolverResult, RandomEngine
from pystreed.binarizer import Binarizer
from pystreed.utils import _dynamic_float_formatter
from typing import Optional
from typing_extensions import Self
import numpy as np
import warnings
import math
import numbers
import sys
import time

class BaseSTreeDSolver(BaseEstimator):

    _parameter_constraints: dict = {
        "optimization_task": [StrOptions({"accuracy", "cost-complex-accuracy", "balanced-accuracy", 
                                          "regression", "cost-complex-regression", 
                                          "piecewise-linear-regression", "simple-linear-regression",
                                          "cost-sensitive", "f1-score", "prescriptive-policy", "instance-cost-sensitive",
                                          "group-fairness", "equality-of-opportunity",
                                          "survival-analysis"})],
        "max_depth": [Interval(numbers.Integral, 0, 20, closed="both")],
        "max_num_nodes": [Interval(numbers.Integral, 0, 1048575, closed="both"), None],
        "min_leaf_node_size": [Interval(numbers.Integral, 1, None, closed="left")],
        "time_limit": [Interval(numbers.Real, 0, None, closed="neither")],
        "cost_complexity": [Interval(numbers.Real, 0, 1, closed="both")],
        "feature_ordering": [StrOptions({"in-order", "gini", "mse"})],
        "upper_bound": [Interval(numbers.Real, 0, None, closed="left")],
        "random_seed": [Interval(numbers.Integral, -1, None, closed="left")],
        "n_thresholds": [Interval(numbers.Integral, 1, None, closed="left")],
        "n_categories": [Interval(numbers.Integral, 2, None, closed="left")],
        "continuous_binarize_strategy": [StrOptions({"tree", "quantile", "uniform"})],
        "max_num_binary_features": [Interval(numbers.Integral, 1, None, closed="left"), None]
    }

    def __init__(self, 
            optimization_task: str,
            max_depth: int = 3,
            max_num_nodes: Optional[int] = None,
            all_trees: bool = False,
            min_leaf_node_size: int = 1,
            time_limit: float = 600,
            cost_complexity : float = 0.01,
            feature_ordering : str = "gini", 
            hyper_tune: bool = False,
            use_branch_caching: bool = False,
            use_dataset_caching: bool = True,
            use_terminal_solver: bool = True,
            use_similarity_lower_bound: bool = True,
            use_upper_bound: bool = True,
            use_lower_bound: bool = True,
            upper_bound: float = 2**31-1,
            verbose: bool = False,
            random_seed: int = 27, 
            continuous_binarize_strategy: str = 'quantile',
            n_thresholds: int = 5,
            n_categories: int = 5, 
            max_num_binary_features: Optional[int] = None):
        """
        Construct a BaseSTreeDSolver

        Parameters:
            optimization_task: the objective used for optimization.
            max_depth: the maximum depth of the tree
            max_num_nodes: the maximum number of branching nodes of the tree
            all_trees: Enable/disable searching for trees in increasing order of size (slower)
            min_leaf_node_size: the minimum number of training instance that should end up in every leaf node
            time_limit: the time limit in seconds for fitting the tree
            cost_complexity: the cost of adding a branch node, expressed as a percentage. E.g., 0.01 means a branching node may be added if it increases the training accuracy by at least 1%.
            feature_ordering: heuristic for the order that features are checked. Default: "gini", alternative: "in-order" or "mse": the order in the given data
            hyper_tune: Use five-fold validation to tune the size of the tree to prevent overfitting
            use_branch_caching: Enable/Disable branch caching (typically the slower caching strategy. May be faster in some scenario's)
            use_dataset_caching: Enable/Disable dataset caching (typically the faster caching strategy)
            use_terminal_solver: Enable/Disable the depth-two solver (Enabled typically results in a large runtime advantage)
            use_similarity_lower_bound: Enable/Disable the similarity lower bound (Enabled typically results in a large runtime advantage)
            use_upper_bound: Enable/Disable the use of upper bounds (Enabled is typically faster)
            use_lower_bound: Enable/Disable the use of lower bounds (Enabled is typically faster)
            upper_bound: Search for a tree better than the provided upper bound
            verbose: Enable/Disable verbose output
            random_seed: the random seed used by the solver (for example when creating folds)
            continuous_binarization_strategy: the strategy used for binarizing continuous features
            n_thresholds: the number of thresholds to use per continuous feature
            n_categories: the number of categories to use per categorical feature
            max_num_binary_features: the maximum number of binary features (selected by random forest feature importance)
        """
        
        self.optimization_task   : str   = optimization_task
        self.max_depth           : int   = max_depth
        self.max_num_nodes       : Optional[int] = max_num_nodes
        self.all_trees           : bool  = all_trees
        self.min_leaf_node_size  : int   = min_leaf_node_size
        self.time_limit          : float = time_limit
        self.cost_complexity     : float = cost_complexity
        self.feature_ordering    : str   = feature_ordering
        self.hyper_tune          : bool  = hyper_tune
        self.use_branch_caching  : bool  = use_branch_caching
        self.use_dataset_caching : bool  = use_dataset_caching
        self.use_terminal_solver : bool  = use_terminal_solver
        self.use_similarity_lower_bound : bool = use_similarity_lower_bound
        self.use_upper_bound     : bool  = use_upper_bound
        self.use_lower_bound     : bool  = use_lower_bound
        self.upper_bound         : float = upper_bound
        self.verbose             : bool  = verbose
        self.random_seed         : int   = random_seed
        self.continuous_binarize_strategy : str = continuous_binarize_strategy
        self.n_thresholds        :int    = n_thresholds
        self.n_categories        : int   = n_categories
        self.max_num_binary_features : Optional[int] = max_num_binary_features
        
        self.fit_result : SolverResult = None
        self._solver = None
        self._label_type = np.int32
        self.binarizer_ : Binarizer = None
        self._prev_data = None
        self._reset_parameters = ["optimization_task", "cost_complexity", "min_leaf_node_size"]

    def _initialize_param_handler(self):
        max_nodes = int(math.pow(2, self.max_depth) - 1)
        max_num_nodes = self.max_num_nodes
        if max_num_nodes is None:
            max_num_nodes = max_nodes
        else:
            if max_num_nodes > max_nodes:
                warnings.warn(f"Maximum number of branching nodes is reduced to {max_nodes} because of the maximum depth {self.max_depth}.", stacklevel=2)
                max_num_nodes  = max_nodes

        self._params = ParameterHandler()
        self._params.optimization_task = self.optimization_task
        self._params.hyper_tune = self.hyper_tune
        self._params.max_depth = self.max_depth
        self._params.max_num_nodes = max_num_nodes
        self._params.all_trees = self.all_trees
        self._params.min_leaf_node_size = self.min_leaf_node_size
        self._params.time_limit = self.time_limit
        if self.optimization_task in ["cost-complex-accuracy", \
                "cost-complex-regression", "piecewise-linear-regression", "simple-linear-regression"]:
            self._params.cost_complexity = self.cost_complexity
        else:
            self._params.cost_complexity = 0.0
        self._params.feature_ordering = self.feature_ordering
        self._params.verbose = self.verbose
        self._params.random_seed = self.random_seed

        self._params.use_branch_caching = self.use_branch_caching
        self._params.use_dataset_caching = self.use_dataset_caching
        self._params.use_terminal_solver = self.use_terminal_solver
        self._params.use_similarity_lower_bound = self.use_similarity_lower_bound
        self._params.use_upper_bound = self.use_upper_bound
        self._params.use_lower_bound = self.use_lower_bound
        self._params.upper_bound = self.upper_bound
    
    def get_solver_params(self) -> ParameterHandler:
        return self._solver._get_parameters()

    def _should_reset_solver(self, X, y, extra_data) -> bool:
        """
        Check if the solver should reset for a new fit. 
        self._reset_parameters contains a list of parameters that, 
        when changed, invalidate the solver's cache
        """
        if self._solver is None: return True
        if self._prev_data is None: return True
        params = self.get_params()
        changed = []
        for key, val in params.items():
            if not hasattr(self._params, key): continue
            cur_val = getattr(self._params, key)
            if cur_val != val:
                changed.append(key)
        if any(s in changed for s in self._reset_parameters):
            return True
        return False        

    def _process_fit_data(self, X, y=None) -> np.ndarray:
        """
        Validate the X and y data before calling fit
        """
        with warnings.catch_warnings():
            warnings.filterwarnings(action="ignore", category=FutureWarning)
            X = self._validate_data(X, ensure_min_samples=2, dtype=np.intc)
            
            if not y is None:
                y = check_array(y, ensure_2d=False, dtype=self._label_type)
                if X.shape[0] != y.shape[0]:
                    raise ValueError('x and y have different number of rows')
                return X, y
            return X
    
    def _process_score_data(self, X, y=None) -> np.ndarray:
        """
        Validate the X and y data before calling score
        """
        with warnings.catch_warnings():
            warnings.filterwarnings(action="ignore", category=FutureWarning)
            X = self._validate_data(X, reset=False, dtype=np.intc)
            
            if not y is None:
                y = check_array(y, ensure_2d=False, dtype=self._label_type)
                if X.shape[0] != y.shape[0]:
                    raise ValueError('x and y have different number of rows')
                return X, y
            return X
    
    def _process_predict_data(self, X) -> np.ndarray:
        """
        Validate the X and y data before calling predict
        """
        with warnings.catch_warnings():
            warnings.filterwarnings(action="ignore", category=FutureWarning)
            return self._validate_data(X, reset=False, dtype=np.intc)

    def _process_extra_data(self, X, extra_data):
        """
        Process the extra data.
        A subclass should override this function if it uses extra data
        """
        if extra_data is None:
            return []
        raise ValueError("extra data is not empty, but the optimization task does not need extra data.")

    def _binarize_data(self, X, y=None, categorical_columns=None, reset=True):
        """
        Binarize the feature data X. If the binarization strategy is "tree", then the label y is also required.
        Binarize the categorical_columns into at most self.n_categories binary features
        Binarize the non-binary, non-categorical features (the continuous features) into self.n_thresholds binary features
        """
        if reset:
            self.binarizer_ = Binarizer(self.continuous_binarize_strategy, self.n_thresholds, self.n_categories, categorical_columns, self.max_num_binary_features)
            self.binarizer_.fit(X, y)
        return self.binarizer_.transform(X)

    def _post_initialize_solver(self):
        """
        Override this function in a subclass if the subclass needs to process
        things after the solver is initialized
        """
        pass

    def fit(self, X, y, extra_data=None, categorical=None) -> Self:
        """
        Fits a STreeD model to the given training data.

        Parameters
        ----------
        x : array-like, shape = (n_samples, n_features)
            Data matrix

        y : array-like, shape = (n_samples)
            Target vector

        extra_data : array-like, shape = (n_samples, n_data_items)
            An array (optional) that represents extra data per instance

        Returns
        ----------
        BaseSTreeDSolver

        Raises
        ----------
        ValueError: If x or y is None or if they have different number of rows.
        """
        # Validate params and data
        self._validate_params()
        X = self._binarize_data(X, y, categorical)
        X, y = self._process_fit_data(X, y)
        extra_data = self._process_extra_data(X, extra_data)
        # Store train data
        self.train_X_ = X
        self.train_y_ = y
        self.train_extra_data_ = extra_data

        if self._should_reset_solver(X, y, extra_data):
            self._initialize_param_handler()
            if self.random_seed == -1:
                self._random_engine = RandomEngine(int(time.time()))
            else:
                self._random_engine = RandomEngine(int(self.random_seed))
            self._solver = initialize_streed_solver(self._params, self._random_engine)
        else:
            self._initialize_param_handler()
            self._solver._update_parameters(self._params)
        self._post_initialize_solver()
                
        self.fit_result = self._solver._solve(X, y, extra_data)
        
        if not self.fit_result.is_feasible():
            warnings.warn("No feasible tree found.", stacklevel=2)
            delattr(self, "fit_result")
        else:
            self.tree_ = self._solver._get_tree(self.fit_result)

        if self.is_fitted() and self.verbose:
            print("Training score: ", self.fit_result.score())
            print("Tree depth: ", self.fit_result.tree_depth(), " \tBranching nodes: ", self.fit_result.tree_nodes())
            if not self.fit_result.is_optimal():
                print("No proof of optimality!")
        
        return self

    def is_fitted(self) -> bool:
        return hasattr(self, "fit_result")

    def predict(self, X, extra_data=None):
        """
        Predicts the target variable for the given input feature data.

        Parameters
        ----------
        x : array-like, shape = (n_samples, n_features)
            Data matrix
        
        extra_data : array-like, shape = (n_samples)
            Extra data (if required)

        Returns
        ----------
        numpy.ndarray: A 1D array that represents the predicted target variable of the test data.
            The i-th element in this array corresponds to the predicted target variable for the i-th instance in `x`.
        """
        check_is_fitted(self, "fit_result")
        X = self._binarize_data(X, reset=False)
        X = self._process_predict_data(X)
        extra_data = self._process_extra_data(X, extra_data)
        return self._solver._predict(self.fit_result, X, extra_data)
    
    def score(self, X, y_true, extra_data=None):
        """
        Computes the score for the given input feature data

        Parameters
        ----------
        x : array-like, shape = (n_samples, n_features)
            Data matrix
        
        y_true : array-like, shape = (n_samples)
            The true labels
        
        extra_data : array-like, shape = (n_samples)
            Extra data (if required)

        Returns
        ----------
        The score
        """
        check_is_fitted(self, "fit_result")
        X = self._binarize_data(X, reset=False)
        X, y_true = self._process_score_data(X, y_true)
        extra_data = self._process_extra_data(X, extra_data)
        self.test_result = self._solver._test_performance(self.fit_result, X, y_true, extra_data)
        return self.test_result.score()
    
    def get_question_length(self, X, y_true, extra_data=None) -> float:
        """
        Computes the average question length for the given input feature data

        Args:
            x : array-like, shape = (n_samples, n_features)
            Data matrix
            y_true : array-like, shape = (n_samples)
            The true labels
            extra_data : array-like, shape = (n_samples)
            Extra data (if required)

        Returns:
            The question length
        """
        check_is_fitted(self, "fit_result")
        X = self._binarize_data(X, reset=False)
        X, y_true = self._process_score_data(X, y_true)
        extra_data = self._process_extra_data(X, extra_data)
        self.test_result = self._solver._test_performance(self.fit_result, X, y_true, extra_data)
        return self.test_result.question_length()

    def get_n_leaves(self) -> int:
        """
        Returns the number of leaf nodes in the fitted tree
        """
        check_is_fitted(self, "fit_result")
        return self.fit_result.tree_nodes() + 1
    
    def get_depth(self) -> int:
        """
        Returns the depth of the fitted tree (a single leaf node is depth zero)
        """
        check_is_fitted(self, "fit_result")
        return self.fit_result.tree_depth()
    
    def get_tree(self):
        """
        Returns the fitted tree
        """
        check_is_fitted(self, "tree_")
        return self.tree_
    
    def _get_label_str(self, label, label_names=None) -> str:
        return str(label) if not isinstance(label, int) or label_names is None else label_names[label]
    
    def _get_predicate_str(self, feature, feature_names=None) -> str:
        if feature_names is None:
            return f"Feature {feature}"
        feature_name = feature_names[feature]
        if " <= " in feature_name:
            feature_name, threshold = feature_name.split(" <= ")
            if len(threshold) >= 3:
                threshold = _dynamic_float_formatter(float(threshold))
            return f"{feature_name} {self.__comparator} {threshold}"
        return feature_name

    def _recursive_print_tree(self, out, node, feature_names, label_names, ind=''):
        if node.is_leaf_node():
            label =  self._get_label_str(node.label, label_names)
            out.write(f"{ind}Label: {label}\n")
        else:
            predicate = self._get_predicate_str(node.feature, feature_names)
            out.write(f"{ind}{predicate} not satisfied\n")
            self._recursive_print_tree(out, node.left_child, feature_names, label_names, ind+"|   ")
            out.write(f"{ind}{predicate} satisfied\n")
            self._recursive_print_tree(out, node.right_child, feature_names, label_names, ind+"|   ")

    def print_tree(self, filename=None, feature_names=None, label_names=None):
        """
        Prints the tree on stdout or writes it to a file (if a filename is given)
        If feature_names is not None, use the names in feature_names for pretty printing
        If label_names is not None, use the names in label_names for pretty printing (only for classification)
        """
        check_is_fitted(self, "tree_")
        
        if feature_names is None and hasattr(self, "feature_names_in_"):
            feature_names = self.feature_names_in_
        self.__comparator = "<="

        if filename is None:
            self._recursive_print_tree(sys.stdout, self.tree_, feature_names, label_names)
        else:
            with open(filename, "w") as fh:
                self._recursive_print_tree(fh, self.tree_, feature_names, label_names)

    def export_dot(self, filename, feature_names=None, label_names=None):
        """
        Write a .dot representation of the tree to filename
        If feature_names is not None, use the names in feature_names for pretty printing
        If label_names is not None, use the names in label_names for pretty printing (only for classification)
        """
        check_is_fitted(self, "tree_")

        if feature_names is None and hasattr(self, "feature_names_in_"):
            feature_names = self.feature_names_in_
        train_data = (self.train_X_, self.train_y_, self.train_extra_data_)
        self.__comparator = "&#8804;" # &#8804; is the <= character

        with open(filename, "w", encoding="utf-8") as fh:
            fh.write("digraph Tree {\n")
            fh.write("node [shape=box, style=\"filled, rounded\", fontname=\"helvetica\", fontsize=\"8\"] ;\n")
            fh.write("edge [fontname=\"helvetica\", fontsize=\"6\"] ;\n")
            self._recursive_export_dot(fh, self.tree_, 0, feature_names, label_names, train_data)
            fh.write("}")
    
    def _export_dot_leaf_node(self, fh, node, node_id, label_names, train_data, color=(200, 200, 200)):
        label =  self._get_label_str(node.label, label_names)
        hex_color = "#{:02x}{:02x}{:02x}".format(*color)
        hex_line_color = "#{:02x}{:02x}{:02x}".format(*[int(0.4 * c) for c in color])
        fh.write(f"{node_id}  [label=\"{label}\", color=\"{hex_line_color}\" fillcolor=\"{hex_color}\"] ;\n")

    def _export_dot_predicate_node(self, fh, node, node_id, feature_names, label_names, train_data):
        predicate = self._get_predicate_str(node.feature, feature_names,)
        fh.write(f"{node_id} [label=\"{predicate}\", color=\"#222222\", fillcolor=\"#EEEEEE\"] ;\n")

    def _recursive_export_dot(self, fh, node, node_id, feature_names, label_names, train_data):
        if node.is_leaf_node():
            self._export_dot_leaf_node(fh, node, node_id, label_names, train_data)
        else:
            self._export_dot_predicate_node(fh, node, node_id, feature_names, label_names, train_data)
        if node_id > 0:
            parent_id = (node_id - 1) // 2
            feature_label = "True" if (node_id % 2) == 0 else "False"
            angle = 45 if feature_label == "True" else -45
            fh.write(f"{parent_id} -> {node_id} [labeldistance=2.5, labelangle={angle}, label=\"{feature_label}\"] ;\n")

        if node.is_branching_node():
            left_data, right_data = self._split(train_data, node.feature)
            self._recursive_export_dot(fh, node.left_child, node_id * 2 + 1, feature_names, label_names, left_data)
            self._recursive_export_dot(fh, node.right_child, node_id * 2 + 2, feature_names, label_names, right_data)

    def _split(self, data, feature):
        x, y, ed = data
        go_left = x[:, feature] == 0
        left_data =  (x[ go_left], y[ go_left], [e for i, e in enumerate(ed) if go_left[i]])
        right_data = (x[~go_left], y[~go_left], [e for i, e in enumerate(ed) if not go_left[i]])
        return left_data, right_data


from pystreed.base import BaseSTreeDSolver
from pystreed.data import ContinuousFeatureData, SimpleContinuousFeatureData
from pystreed.binarizer import get_column_types
from sklearn.base import RegressorMixin
from sklearn.utils.validation import check_is_fitted
from sklearn.utils._param_validation import Interval, StrOptions
from typing import Optional
import numpy as np
import pandas as pd
import numbers
from pathlib import Path
import os

class STreeDRegressor(BaseSTreeDSolver, RegressorMixin):

    _parameter_constraints: dict = {**BaseSTreeDSolver._parameter_constraints, 
        "feature_ordering": [StrOptions({"in-order", "mse"})],
        "regression_lower_bound": [StrOptions({"equivalent", "kmeans"})]
    }

    def __init__(self,
                 optimization_task : str = "cost-complex-regression",
                 max_depth : int = 3,
                 max_num_nodes : Optional[int] = None,
                 min_leaf_node_size : int = 1,
                 time_limit : float = 600,
                 feature_ordering : str = "mse", 
                 hyper_tune : bool = False,
                 cost_complexity : float = 0.01,
                 regression_lower_bound : str = "kmeans",
                 use_branch_caching: bool = True,
                 use_dataset_caching: bool = False,
                 use_terminal_solver: bool = True,
                 use_similarity_lower_bound: bool = True,
                 use_upper_bound: bool = True,
                 use_lower_bound: bool = True,
                 use_task_lower_bound: bool = True,
                 upper_bound: float = 2**31 - 1,
                 verbose : bool = False,
                 random_seed: int = 27, 
                 continuous_binarize_strategy: str = 'quantile',
                 n_thresholds: int = 5,
                 n_categories: int = 5,
                 max_num_binary_features: int = None):
        """
        Construct a STreedRegressor

        Parameters:
            optimization_task: the objective used for optimization. Default = cost-complex-regression
            max_depth: the maximum depth of the tree
            max_num_nodes: the maximum number of branching nodes of the tree
            min_leaf_node_size: the minimum number of training instance that should end up in every leaf node
            time_limit: the time limit in seconds for fitting the tree
            feature_ordering: heuristic for the order that features are checked. Default: "mse", alternative: "in-order": the order in the given data            
            hyper_tune: Use five-fold validation to tune the size of the tree to prevent overfitting
            cost_complexity: the cost of adding a branch node, expressed as a percentage. E.g., 0.01 means a branching node may be added if it increases the training accuracy by at least 1%.
                only used when optimization_task == "cost-complex-regression'
            regression_lower_bound: the lower bound used by the cost-complex-regression task: kmeans or equivalent
            use_branch_caching: Enable/Disable branch caching (typically the slower caching strategy. May be faster in some scenario's)
            use_dataset_caching: Enable/Disable dataset caching (typically the faster caching strategy)
            use_terminal_solver: Enable/Disable the depth-two solver (Enabled typically results in a large runtime advantage)
            use_similarity_lower_bound: Enable/Disable the similarity lower bound (Enabled typically results in a large runtime advantage)
            use_upper_bound: Enable/Disable the use of upper bounds (Enabled is typically faster)
            use_lower_bound: Enable/Disable the use of lower bounds (Enabled is typically faster)
            use_task_lower_bound: Enable/Disable the kmeans/equivalent lower bound for cost-complex-regression (typically faster)
            upper_bound: Search for a tree better than the provided upper bound
            verbose: Enable/Disable verbose output
            random_seed: the random seed used by the solver (for example when creating folds)
            continuous_binarization_strategy: the strategy used for binarizing continuous features
            n_thresholds: the number of thresholds to use per continuous feature
            n_categories: the number of categories to use per categorical feature
            max_num_binary_features: the maximum number of binary features (selected by random forest feature importance)
        """
        self._extra_permitted_params = ["regression_lower_bound"]
        if not optimization_task in ["regression", "cost-complex-regression"]:
            raise ValueError(f"Invalid value for optimization_task: {optimization_task}")
        BaseSTreeDSolver.__init__(self, optimization_task,
            max_depth=max_depth,
            max_num_nodes=max_num_nodes,
            min_leaf_node_size=min_leaf_node_size,
            time_limit=time_limit,
            cost_complexity=cost_complexity,
            feature_ordering=feature_ordering,
            hyper_tune = hyper_tune,
            use_branch_caching=use_branch_caching,
            use_dataset_caching=use_dataset_caching,
            use_terminal_solver=use_terminal_solver,
            use_similarity_lower_bound=use_similarity_lower_bound,
            use_upper_bound=use_upper_bound,
            use_lower_bound=use_lower_bound,
            upper_bound=upper_bound,
            verbose=verbose,
            random_seed=random_seed,
            continuous_binarize_strategy=continuous_binarize_strategy,
            n_thresholds=n_thresholds,
            n_categories=n_categories,
            max_num_binary_features=max_num_binary_features)
        self._label_type = np.double
        self.regression_lower_bound = regression_lower_bound
        self.use_task_lower_bound = use_task_lower_bound

    def _initialize_param_handler(self):
        super()._initialize_param_handler()
        self._params.regression_lower_bound = self.regression_lower_bound
        self._params.use_task_lower_bound = self.use_task_lower_bound
        return self._params

class STreeDPiecewiseLinearRegressor(BaseSTreeDSolver):

    _parameter_constraints: dict = {**BaseSTreeDSolver._parameter_constraints, 
        "lasso_penalty": [Interval(numbers.Real, 0, 1e12, closed="both")],
        "ridge_penalty": [Interval(numbers.Real, 0, 1e12, closed="both")],
    }

    def __init__(self,
                 simple : bool = False,
                 max_depth : int = 3,
                 max_num_nodes : Optional[int] = None,
                 min_leaf_node_size : int = 1,
                 time_limit : float = 600,
                 feature_ordering : str = "mse", 
                 hyper_tune : bool = False,
                 cost_complexity : float = 0.01,
                 lasso_penalty : float = 0.0,
                 ridge_penalty : float = 0.0,
                 use_branch_caching: bool = False,
                 use_dataset_caching: bool = True,
                 use_similarity_lower_bound: bool = True,
                 use_upper_bound: bool = True,
                 use_lower_bound: bool = True,
                 upper_bound: float = 2**31 - 1,
                 verbose : bool = False,
                 random_seed: int = 27, 
                 continuous_binarize_strategy: str = 'quantile',
                 n_thresholds: int = 5,
                 n_categories: int = 5,
                 max_num_binary_features: int = None):
        """
        Construct a STreeDPiecewiseLinearRegressor

        Parameters:
            simple: set to True to fit a simple linear regression model. False otherwise
            max_depth: the maximum depth of the tree
            max_num_nodes: the maximum number of branching nodes of the tree
            min_leaf_node_size: the minimum number of training instance that should end up in every leaf node
            time_limit: the time limit in seconds for fitting the tree
            feature_ordering: heuristic for the order that features are checked. Default: "mse", alternative: "in-order": the order in the given data            
            hyper_tune: Use five-fold validation to tune the size of the tree to prevent overfitting
            cost_complexity: the cost of adding a branch node, expressed as a percentage. E.g., 0.01 means a branching node may be added if it increases the training accuracy by at least 1%.
            lasso_penalty: the amount of lasso penalization used in each leaf node
            ridge_penalty: the amount of ridge penalization in each leaf node
            use_branch_caching: Enable/Disable branch caching (typically the slower caching strategy. May be faster in some scenario's)
            use_dataset_caching: Enable/Disable dataset caching (typically the faster caching strategy)
            use_similarity_lower_bound: Enable/Disable the similarity lower bound (Enabled typically results in a large runtime advantage)
            use_upper_bound: Enable/Disable the use of upper bounds (Enabled is typically faster)
            use_lower_bound: Enable/Disable the use of lower bounds (Enabled is typically faster)
            upper_bound: Search for a tree better than the provided upper bound
            verbose: Enable/Disable verbose output
            random_seed: the random seed used by the solver (for example when creating folds)
            continuous_binarization_strategy: the strategy used for binarizing continuous features
            n_thresholds: the number of thresholds to use per continuous feature
            n_categories: the number of categories to use per categorical feature
            max_num_binary_features: the maximum number of binary features (selected by random forest feature importance)
        """
        BaseSTreeDSolver.__init__(self, "simple-linear-regression" if simple else "piecewise-linear-regression",
            max_depth=max_depth,
            max_num_nodes=max_num_nodes,
            min_leaf_node_size=min_leaf_node_size,
            time_limit=time_limit,
            cost_complexity=cost_complexity,
            feature_ordering=feature_ordering,
            hyper_tune = hyper_tune,
            use_branch_caching=use_branch_caching,
            use_dataset_caching=use_dataset_caching,
            use_terminal_solver=False,
            use_similarity_lower_bound=use_similarity_lower_bound,
            use_upper_bound=use_upper_bound,
            use_lower_bound=use_lower_bound,
            upper_bound=upper_bound,
            verbose=verbose,
            random_seed=random_seed,
            continuous_binarize_strategy=continuous_binarize_strategy,
            n_thresholds=n_thresholds,
            n_categories=n_categories,
            max_num_binary_features=max_num_binary_features)
        self.simple = simple
        self._label_type = np.double
        self.label_name_ = None
        self.continuous_columns_ = None
        self.n_continuous_columns_ = 0
        self.lasso_penalty = lasso_penalty
        self.ridge_penalty = ridge_penalty
        self._reset_parameters.append("lasso_penalty")
        self._reset_parameters.append("ridge_penalty")

    def _initialize_param_handler(self):
        super()._initialize_param_handler()
        self._params.lasso_penalty = self.lasso_penalty
        self._params.ridge_penalty = self.ridge_penalty
        n_vars = 1 if self.simple else self.n_continuous_columns_
        if self.min_leaf_node_size < 5 * n_vars:
            if self.verbose:
                print(f"Updating the minimum leaf node size to {5 * n_vars}.")
            self._params.min_leaf_node_size = 5 * n_vars
        return self._params

    def _process_extra_data(self, X, extra_data):
        return extra_data

    def _get_continuous_columns(self, X, continuous_columns):
        if continuous_columns is None:
            continuous_columns = self.continuous_columns_
        if continuous_columns is None:
            _, _, continuous_columns = get_column_types(X, self.categorical_columns_)
            if len(continuous_columns) == 0:
                raise ValueError("Provide continuous columns.")
        if isinstance(continuous_columns, list): # column names
            self.continuous_columns_ = continuous_columns
            if isinstance(X, np.ndarray):
                cont_cols = X[:, continuous_columns]
            elif isinstance(X, pd.DataFrame):
                cont_cols = X[continuous_columns].values
            else:
                raise ValueError(f"X should be a numpy array or a pandas data frame. X.type = {type(X)}")
        elif isinstance(continuous_columns, np.ndarray):
            cont_cols = continuous_columns
        elif isinstance(continuous_columns, pd.DataFrame):
            cont_cols = continuous_columns.values
        else:
            raise ValueError(f"X should be a numpy array or a pandas data frame. X.type = {type(X)}")
        self.n_continuous_columns_ = cont_cols.shape[1]
        if self.simple:
            extra_data = [SimpleContinuousFeatureData(list(row)) for row in cont_cols]
        else:
            extra_data = [ContinuousFeatureData(list(row)) for row in cont_cols]
        return extra_data

    def fit(self, X, y, continuous_columns=None, categorical_columns=None):
        """
        Fits a STreeD Regression model to the given training data.

        Args:
            x : array-like, shape = (n_samples, n_features)
            Data matrix

            y : array-like, shape = (n_samples)
            Target vector

            extra_data : array-like, shape = (n_samples, n_data_items)
            An array (optional) that represents extra data per instance

            continuous_columns : array-like
            List of columns that are continuous (will be used for the regression)
            If None, continuous columns are automatically detected

            categorical_columns : array-like, 
            List of column names that are categorical

        Returns:
            BaseSTreeDSolver

        Raises:
            ValueError: If x or y is None or if they have different number of rows.
        """
        self.categorical_columns_ = categorical_columns
        extra_data = self._get_continuous_columns(X, continuous_columns)
        if isinstance(y, pd.Series): self.label_name_ = y.name
        return super().fit(X, y, extra_data, categorical_columns)

    def predict(self, X, continuous_columns=None):
        check_is_fitted(self, "fit_result")
        extra_data = self._get_continuous_columns(X, continuous_columns)
        return super().predict(X, extra_data)
    
    def score(self, X, y_true, continuous_columns=None):
        check_is_fitted(self, "fit_result")
        extra_data = self._get_continuous_columns(X, continuous_columns)
        return super().score(X, y_true, extra_data)
    
    def _export_dot_leaf_node(self, fh, node, node_id, label_names, train_data):
        try:
            x, y, ed = train_data
            import matplotlib.pyplot as plt
            plt.rc('axes', labelsize=6)
            plt.rc('xtick', labelsize=6)
            plt.rc('ytick', labelsize=6)
            linear_model = node.label
            # plot on the axis with the highest difference in predicted values
            nCF = len(linear_model.coefficients)
            x_index = 0
            max_diff = 0
            for index in range(nCF):
                xj = [e.feature_data[index] for e in ed]
                yhat = [_x * linear_model.coefficients[index] for _x in xj]
                diff = max(yhat) - min(yhat)
                if diff > max_diff:
                    x_index = index
                    max_diff = diff
            intercept = linear_model.intercept
            coef = linear_model.coefficients[x_index]
            # add the mean for other features multiplied by the coefficient to the intercept
            for index in range(nCF):
                if index == x_index: continue
                xj = [e.feature_data[index] for e in ed]
                intercept += linear_model.coefficients[index] * np.mean(xj)
            xj = [e.feature_data[x_index] for e in ed]
            yhat = [intercept + _x * coef for _x in xj]
            xj, yhat, y = list(zip(*sorted(zip(xj, yhat, y), key=lambda v: v[0])))
            # plot the figure 
            plt.figure(figsize=(2,2))            
            plt.grid(zorder=-2.0)
            plt.plot(xj, yhat, zorder=3)
            plt.scatter(xj, y, color='red', s=2, zorder=2)
            
            if not label_names is None and isinstance(label_names, str): plt.ylabel()
            elif not self.label_name_ is None: plt.ylabel("(Predicted) Value" if self.label_name_==0 else self.label_name_)
            if isinstance(self.continuous_columns_[x_index], int):
                plt.xlabel(f"Feature {self.continuous_columns_[x_index]}")  
            else:
                plt.xlabel(f"{self.continuous_columns_[x_index]}")
            filename = self.__export_directory / f"leaf_reg_{node_id}.png"
            if not os.path.exists(self.__export_directory):
                os.makedirs(self.__export_directory)
            plt.savefig(filename, bbox_inches="tight", dpi=600)
            fh.write(f"{node_id}  [label=\"\", image=\"{filename}\", width=2, height=2, fixedsize=true, shape=none] ;\n")
        except:
            return super()._export_dot_leaf_node(fh, node, node_id, label_names, train_data)

    def export_dot(self, filename, feature_names=None, label_names=None):
        """
        Export the tree as a .dot file (for plotting)
        """
        self.__export_directory = Path(filename).parent.absolute() / "leaf-distributions"
        return super().export_dot(filename, feature_names, label_names)
    



from pystreed.base import BaseSTreeDSolver
from typing import Optional
from sklearn.base import ClassifierMixin
from sklearn.utils._param_validation import Interval, StrOptions
from sklearn.utils.validation import check_is_fitted
from pystreed.utils import _color_brew
import numpy as np
import numbers
import warnings

class STreeDClassifier(BaseSTreeDSolver, ClassifierMixin):
    """
    STreeDClassifier returns optimal classification trees.
    It supports several objectives, as specified by the optimization task parameter
    """

    _parameter_constraints: dict = {**BaseSTreeDSolver._parameter_constraints, 
        "feature_ordering": [StrOptions({"in-order", "gini"})]
    }

    def __init__(self, 
                 optimization_task : str = "accuracy",
                 max_depth : int = 3,
                 max_num_nodes : Optional[int] = None,
                 all_trees : bool = False,
                 min_leaf_node_size: int = 1,
                 time_limit : float = 600,
                 cost_complexity : float = 0.01,
                 feature_ordering : str = "gini", 
                 hyper_tune: bool = False,
                 use_branch_caching: bool = False,
                 use_dataset_caching: bool = True,
                 use_terminal_solver: bool = True,
                 use_similarity_lower_bound: bool = True,
                 use_upper_bound: bool = True,
                 use_lower_bound: bool = True,
                 use_task_lower_bound: bool = True,
                 upper_bound: float = 2**31 -1,
                 verbose : bool = False,
                 random_seed: int = 27, 
                 continuous_binarize_strategy: str = 'quantile',
                 n_thresholds: int = 5,
                 n_categories: int = 5, 
                 max_num_binary_features: int = None):
        """
        Construct a STreeDClassifier

        Parameters:
            optimization_task: the objective used for optimization. Default = accuracy
            max_depth: the maximum depth of the tree
            max_num_nodes: the maximum number of branching nodes of the tree
            all_trees: Enable/disable searching for trees in increasing order of size (slower)
            min_leaf_node_size: the minimum number of training instance that should end up in every leaf node
            time_limit: the time limit in seconds for fitting the tree
            cost_complexity: the cost of adding a branch node, expressed as a percentage. E.g., 0.01 means a branching node may be added if it increases the training accuracy by at least 1%.
                only used when optimization_task == "cost-complex-accuracy'
            feature_ordering: heuristic for the order that features are checked. Default: "gini", alternative: "in-order": the order in the given data
            hyper_tune: Use five-fold validation to tune the size of the tree to prevent overfitting
            use_branch_caching: Enable/Disable branch caching (typically the slower caching strategy. May be faster in some scenario's)
            use_dataset_caching: Enable/Disable dataset caching (typically the faster caching strategy)
            use_terminal_solver: Enable/Disable the depth-two solver (Enabled typically results in a large runtime advantage)
            use_similarity_lower_bound: Enable/Disable the similarity lower bound (Enabled typically results in a large runtime advantage)
            use_upper_bound: Enable/Disable the use of upper bounds (Enabled is typically faster)
            use_lower_bound: Enable/Disable the use of lower bounds (Enabled is typically faster)
            use_task_lower_bound: Enable/Disable the equivalent point lower bound for cost-complex-accuracy
            upper_bound: Search for a tree better than the provided upper bound
            verbose: Enable/Disable verbose output
            random_seed: the random seed used by the solver (for example when creating folds)
            continuous_binarization_strategy: the strategy used for binarizing continuous features
            n_thresholds: the number of thresholds to use per continuous feature
            n_categories: the number of categories to use per categorical feature
            max_num_binary_features: the maximum number of binary features (selected by random forest feature importance)
        """
        if not optimization_task in ["accuracy", "cost-complex-accuracy", "balanced-accuracy", "f1-score"]:
            raise ValueError(f"Invalid value for optimization_task: {optimization_task}")
        BaseSTreeDSolver.__init__(self, optimization_task, 
            max_depth=max_depth,
            max_num_nodes=max_num_nodes,
            all_trees = all_trees,
            min_leaf_node_size=min_leaf_node_size,
            time_limit=time_limit,
            cost_complexity=cost_complexity,
            feature_ordering=feature_ordering,
            hyper_tune=hyper_tune,
            use_branch_caching=use_branch_caching,
            use_dataset_caching=use_dataset_caching,
            use_terminal_solver=use_terminal_solver,
            use_similarity_lower_bound=use_similarity_lower_bound,
            use_upper_bound=use_upper_bound,
            use_lower_bound=use_lower_bound,
            upper_bound=upper_bound,
            verbose=verbose,
            random_seed=random_seed,
            continuous_binarize_strategy=continuous_binarize_strategy,
            n_thresholds=n_thresholds,
            n_categories=n_categories,
            max_num_binary_features=max_num_binary_features)
        self.use_task_lower_bound = use_task_lower_bound
        if optimization_task == "f1-score" and upper_bound != 2**31-1:
            warnings.warn(f"upper_bound parameter is ignored for f1-score", stacklevel=2)
        
    def _initialize_param_handler(self):
        super()._initialize_param_handler()
        self._params.use_task_lower_bound = self.use_task_lower_bound
        return self._params

    def fit(self, X, y, extra_data=None, categorical=None):
        """
        Fits a STreeD Classification model to the given training data.

        Args:
            x : array-like, shape = (n_samples, n_features)
            Data matrix

            y : array-like, shape = (n_samples)
            Target vector

            extra_data : array-like, shape = (n_samples, n_data_items)
            An array (optional) that represents extra data per instance

            categorical : array-like, 
            List of column names that are categorical

        Returns:
            BaseSTreeDSolver

        Raises:
            ValueError: If x or y is None or if they have different number of rows.
        """
        self.classes_ = list(sorted(np.unique(y)))
        self.n_classes_ = len(self.classes_)
        return super().fit(X, y, extra_data, categorical)

    def predict_proba(self, X, extra_data=None):
        """
        Predicts the probabilities of the target class for the given input feature data.

        Args:
            X : array-like, shape = (n_samples, n_features)
            Data matrix
            extra_data : array-like, shape = (n_samples)
            Extra data (if required)

        Returns:
            numpy.ndarray: A 2D array that represents the predicted class probabilities of the test data.
                The i-j-th element in this array corresponds to the predicted class probablity for the j-th class of the i-th instance in `X`.
        """
        check_is_fitted(self, "fit_result")
        X = self._binarize_data(X, reset=False)
        X = self._process_predict_data(X)
        extra_data = self._process_extra_data(X, extra_data)
        probabilities = np.zeros((len(X), self.n_classes_))
        train_data = (self.train_X_, self.train_y_)
        self._recursive_predict_proba(self.tree_, probabilities, np.array(range(0, len(X))), X, train_data)
        # Check that all rows sum to proability 1 (account for floating errors)
        assert (probabilities.sum(axis=1).min() >= 1-1e-4)
        return probabilities
    
    def _recursive_predict_proba(self, tree, probabilities, indices, X, train_data):
        train_X = train_data[0]
        train_y = train_data[1]
        if tree.is_leaf_node():
            n = len(train_y)
            assert(n > 0)
            all_counts = np.zeros(self.n_classes_)
            unique, counts = np.unique(train_y, return_counts=True)
            for label, count in zip(unique, counts):
                all_counts[label] = count
            probs = all_counts / n
            probabilities[indices] = probs
        else:
            indices_left  = np.intersect1d(np.argwhere(~X[:, tree.feature]), indices)
            indices_right = np.intersect1d(np.argwhere( X[:, tree.feature]), indices)
            sel = train_X[:, tree.feature].astype(bool)
            train_data_left  = (train_X[~sel, :], train_y[~sel])
            train_data_right = (train_X[ sel, :], train_y[ sel])
            self._recursive_predict_proba(tree.left_child,  probabilities, indices_left,  X, train_data_left)
            self._recursive_predict_proba(tree.right_child, probabilities, indices_right, X, train_data_right)

    def _export_dot_leaf_node(self, fh, node, node_id, label_names, train_data):
        if not hasattr(self, "_colors"):
            self._colors = _color_brew(self.n_classes_)
        color = self._colors[node.label]
        return super()._export_dot_leaf_node(fh, node, node_id, label_names, train_data, color=color)


import numpy as np
import math

def _color_brew(n):
    """Generate n colors with equally spaced hues.
    From: https://github.com/scikit-learn/scikit-learn/blob/main/sklearn/tree/_export.py

    Parameters
    ----------
    n : int
        The number of colors required.

    Returns
    -------
    color_list : list, length n
        List of n tuples of form (R, G, B) being the components of each color.
    """
    color_list = []

    # Initialize saturation & value; calculate chroma & value shift
    s, v = 0.75, 0.9
    c = s * v
    m = v - c

    for h in np.arange(25, 385, 360.0 / n).astype(int):
        # Calculate some intermediate values
        h_bar = h / 60.0
        x = c * (1 - abs((h_bar % 2) - 1))
        # Initialize RGB with same hue & chroma as our color
        rgb = [
            (c, x, 0),
            (x, c, 0),
            (0, c, x),
            (0, x, c),
            (x, 0, c),
            (c, 0, x),
            (c, x, 0),
        ]
        r, g, b = rgb[int(h_bar)]
        # Shift the initial RGB values to match value and store
        rgb = [(int(255 * (r + m))), (int(255 * (g + m))), (int(255 * (b + m)))]
        color_list.append(rgb)

    return color_list

def _dynamic_float_formatter(t):
    if int(t) == t:
        return str(t)
    if t % 1 == 0:
        return str(t)
    if math.log10(abs(t)) >= 6 or math.log10(abs(t)) <= -4:
        return f"{t:.2e}"
    if math.log10(abs(t)) >= 2:
        return f"{t:.2f}".rstrip('0').rstrip('.')
    return f"{t:f}".rstrip('0').rstrip('.')


from pystreed.base import BaseSTreeDSolver
from typing import Optional
from pystreed.utils import _color_brew
import numpy as np
from pystreed.data import CostSpecifier

class STreeDCostSensitiveClassifier(BaseSTreeDSolver):

    def __init__(self, 
                 max_depth : int = 3,
                 max_num_nodes : Optional[int] = None,
                 min_leaf_node_size: int = 1,
                 time_limit : float = 600,
                 feature_ordering : str = "gini", 
                 hyper_tune: bool = False,
                 use_branch_caching: bool = True,
                 use_terminal_solver: bool = True,
                 use_upper_bound: bool = True,
                 use_lower_bound: bool = True,
                 upper_bound: float = 2**31-1,
                 verbose : bool = False,
                 random_seed: int = 27):
        """
        Construct a STreeDCostSensitiveClassifier

        Parameters:
            max_depth: the maximum depth of the tree
            max_num_nodes: the maximum number of branching nodes of the tree
            min_leaf_node_size: the minimum number of training instance that should end up in every leaf node
            time_limit: the time limit in seconds for fitting the tree
            feature_ordering: heuristic for the order that features are checked. Default: "gini", alternative: "in-order": the order in the given data
            hyper_tune: Use five-fold validation to tune the size of the tree to prevent overfitting
            use_branch_caching: Enable/Disable branch caching
            use_terminal_solver: Enable/Disable the depth-two solver (Enabled typically results in a large runtime advantage)
            use_upper_bound: Enable/Disable the use of upper bounds (Enabled is typically faster)
            use_lower_bound: Enable/Disable the use of lower bounds (Enabled is typically faster)
            upper_bound: Search for a tree better than the provided upper bound
            verbose: Enable/Disable verbose output
            random_seed: the random seed used by the solver (for example when creating folds)
        """
        BaseSTreeDSolver.__init__(self, "cost-sensitive", 
            max_depth=max_depth,
            max_num_nodes=max_num_nodes,
            min_leaf_node_size=min_leaf_node_size,
            time_limit=time_limit,
            cost_complexity=0,
            feature_ordering=feature_ordering,
            hyper_tune=hyper_tune,
            use_branch_caching=use_branch_caching,
            use_dataset_caching=False,
            use_terminal_solver=use_terminal_solver,
            use_similarity_lower_bound=False,
            use_upper_bound=use_upper_bound,
            use_lower_bound=use_lower_bound,
            upper_bound=upper_bound,
            verbose=verbose,
            random_seed=random_seed)
        
    def _post_initialize_solver(self):
        super()._post_initialize_solver()
        self._solver.specify_costs(self.cost_specifier_)

    def _binarize_data(self, X, y=None, categorical_columns=None, reset=True):
        return X

    def fit(self, X, y, cost_specifier : CostSpecifier, categorical=None):
        """
        Fits a STreeD model to the given training data.

        Args:
            X : array-like, shape = (n_samples, n_features)
            Data matrix

            y : array-like, shape = (n_samples)
            Target vector

            cost_specifier : CostSpecifier
            An object that describes the misclassification matrix and the feature cost function

            categorical : array-like, 
            List of column names that are categorical

        Returns:
            BaseSTreeDSolver

        Raises:
            ValueError: If x or y is None or if they have different number of rows.
        """
        self.n_classes_ = len(np.unique(y))
        self.cost_specifier_ = cost_specifier
        return super().fit(X, y, categorical)
        
    def _export_dot_leaf_node(self, fh, node, node_id, label_names, train_data):
        if not hasattr(self, "_colors"):
            self._colors = _color_brew(self.n_classes_)
        color = self._colors[node.label]
        return super()._export_dot_leaf_node(fh, node, node_id, label_names, train_data, color=color)


from pystreed.base import BaseSTreeDSolver
from typing import Optional, Union
from sklearn.utils._param_validation import Interval
from pystreed.utils import _color_brew
import numpy as np
import pandas as pd
import numbers

class STreeDGroupFairnessClassifier(BaseSTreeDSolver):

    _parameter_constraints: dict = {**BaseSTreeDSolver._parameter_constraints, 
        "discrimination_limit": [Interval(numbers.Real, 0, 1, closed="both")]
    }

    def __init__(self, 
                 optimization_task : str = 'group-fairness',
                 sensitive_feature : Union[int, str] = 0,
                 discrimination_limit : float = 0.05,
                 max_depth : int = 3,
                 max_num_nodes : Optional[int] = None,
                 min_leaf_node_size: int = 1,
                 time_limit : float = 600,
                 feature_ordering : str = "gini", 
                 hyper_tune: bool = False,
                 use_branch_caching: bool = True,
                 use_terminal_solver: bool = True,
                 use_similarity_lower_bound: bool = True,
                 use_upper_bound: bool = True,
                 use_lower_bound: bool = True,
                 verbose : bool = False,
                 random_seed: int = 27, 
                 continuous_binarize_strategy: str = 'quantile',
                 n_thresholds: int = 5,
                 n_categories: int = 5):
        """
        Construct a STreeDGroupFairnessClassifier

        Parameters:
            optimization_task: the objective used for optimization.
            sensitive_feature: str or int, the name or index of the sensitive feature (default is index 0)
            max_depth: the maximum depth of the tree
            max_num_nodes: the maximum number of branching nodes of the tree
            min_leaf_node_size: the minimum number of training instance that should end up in every leaf node
            time_limit: the time limit in seconds for fitting the tree
            feature_ordering: heuristic for the order that features are checked. Default: "gini", alternative: "in-order": the order in the given data
            hyper_tune: Use five-fold validation to tune the size of the tree to prevent overfitting
            use_branch_caching: Enable/Disable branch caching.
            use_terminal_solver: Enable/Disable the depth-two solver (Enabled typically results in a large runtime advantage)
            use_similarity_lower_bound: Enable/Disable the similarity lower bound (Enabled typically results in a large runtime advantage)
            use_upper_bound: Enable/Disable the use of upper bounds (Enabled is typically faster)
            use_lower_bound: Enable/Disable the use of lower bounds (Enabled is typically faster)
            verbose: Enable/Disable verbose output
            random_seed: the random seed used by the solver (for example when creating folds)
            continuous_binarization_strategy: the strategy used for binarizing continuous features
            n_thresholds: the number of thresholds to use per continuous feature
            n_categories: the number of categories to use per categorical feature
        """

        if not optimization_task in ["group-fairness", "equality-of-opportunity"]:
            raise ValueError(f"Invalid value for optimization_task: {optimization_task}")
        BaseSTreeDSolver.__init__(self, optimization_task, 
            max_depth=max_depth,
            max_num_nodes=max_num_nodes,
            min_leaf_node_size=min_leaf_node_size,
            time_limit=time_limit,
            cost_complexity=0,
            feature_ordering=feature_ordering,
            hyper_tune=hyper_tune,
            use_branch_caching=use_branch_caching,
            use_dataset_caching=False,
            use_terminal_solver=use_terminal_solver,
            use_similarity_lower_bound=use_similarity_lower_bound,
            use_upper_bound=use_upper_bound,
            use_lower_bound=use_lower_bound,
            verbose=verbose,
            random_seed=random_seed,
            continuous_binarize_strategy=continuous_binarize_strategy,
            n_thresholds=n_thresholds,
            n_categories=n_categories)
        self.discrimination_limit = discrimination_limit
        self.sensitive_feature = sensitive_feature
        
    def _initialize_param_handler(self):
        super()._initialize_param_handler()
        self._params.discrimination_limit = self.discrimination_limit
        return self._params

    def _move_sensitive_feature_first(self, X):
        if isinstance(self.sensitive_feature, int):
            if self.sensitive_feature == 0: return X
            if isinstance(X, pd.DataFrame):
                sensitive_feature = X.columns[self.sensitive_feature]
                sensitive_column = X.pop(sensitive_feature) 
                X.insert(0, sensitive_feature, sensitive_column) 
                return X
            return X[:, [self.sensitive_feature, *range(0,self.sensitive_feature), *range(self.sensitive_feature+1, X.shape[1])]]

        elif isinstance(self.sensitive_feature, str):
            # assume pandas dataframe
            if not self.sensitive_feature in X.columns:
                 raise ValueError(f"Column {self.sensitive_feature} is not part of the data.")
            sensitive_column = X.pop(self.sensitive_feature) 
            X.insert(0, self.sensitive_feature, sensitive_column)
            return X
        else:
            raise ValueError(f"The sensitive feature column should be either an index or a string, but is {self.sensitive_feature} of type {type(self.sensitive_feature)}.")

    def fit(self, X, y, categorical=None):
        """
        Fits a STreeD model to the given training data.

        Args:
            x : array-like, shape = (n_samples, n_features)
            Data matrix

            y : array-like, shape = (n_samples)
            Target vector

            categorical : array-like, 
            List of column names that are categorical

        Returns:
            BaseSTreeDSolver

        Raises:
            ValueError: If x or y is None or if they have different number of rows.
        """
        self.n_classes_ = len(np.unique(y))
        X = self._move_sensitive_feature_first(X)
        return super().fit(X, y, categorical)
    
    def predict(self, X):
        """
        Predicts the target variable for the given input feature data.

        Args:
            x : array-like, shape = (n_samples, n_features)
            Data matrix

        Returns:
            numpy.ndarray: A 1D array that represents the predicted target variable of the test data.
                The i-th element in this array corresponds to the predicted target variable for the i-th instance in `x`.
        """
        X = self._move_sensitive_feature_first(X)
        return super().predict(X)
    
    def score(self, X, y_true):
        """
        Computes the score for the given input feature data

        Args:
            x : array-like, shape = (n_samples, n_features)
            Data matrix
            y_true : array-like, shape = (n_samples)
            The true labels

        Returns:
            The score
        """
        X = self._move_sensitive_feature_first(X)
        return super().score(X, y_true)
        
    def _export_dot_leaf_node(self, fh, node, node_id, label_names, train_data):
        if not hasattr(self, "_colors"):
            self._colors = _color_brew(self.n_classes_)
        color = self._colors[node.label]
        return super()._export_dot_leaf_node(fh, node, node_id, label_names, train_data, color=color)


from pystreed.base import *
from pystreed.classification import STreeDClassifier
from pystreed.cost_sensitive_classification import STreeDCostSensitiveClassifier
from pystreed.regression import STreeDRegressor, STreeDPiecewiseLinearRegressor
from pystreed.survival_analysis import STreeDSurvivalAnalysis
from pystreed.prescriptive_policy_generation import STreeDPrescriptivePolicyGenerator
from pystreed.group_fair import STreeDGroupFairnessClassifier
from pystreed.instance_cost_sensitive_classification import STreeDInstanceCostSensitiveClassifier
from pystreed.data import *


# Available at setup time due to pyproject.toml
from pybind11.setup_helpers import Pybind11Extension, build_ext
from setuptools import setup
from glob import glob

# Define package metadata
package_name = 'pystreed'
extension_name = 'cstreed'
__version__ = "1.3.6"

ext_modules = [
    Pybind11Extension(package_name + '.' + extension_name,
        sorted(glob("src/**/*.cpp", recursive = True)),
        include_dirs = ["include"],
        define_macros = [('VERSION_INFO', __version__)], # passing in the version to the compiled code
        language='c++',
        cxx_std=17
    )
]

setup(
    name=package_name,
    version=__version__,
    ext_modules=ext_modules,
    dev_requires=['pytest'],
    install_requires=['pandas', 'numpy'],
    # Add read me as long description and replace relative example links with absolute links
    long_description=open('README.md').read().replace("](examples", "](https://github.com/AlgTUDelft/pystreed/blob/main/examples"),
    long_description_content_type='text/markdown'
)


from pystreed.data import CostVector
from pystreed.base import BaseSTreeDSolver
from sklearn.utils._param_validation import Interval, StrOptions
from typing import Optional
from pystreed.utils import _color_brew, _dynamic_float_formatter
import numpy as np
import pandas as pd
import os
from pathlib import Path

class STreeDInstanceCostSensitiveClassifier(BaseSTreeDSolver):

    _parameter_constraints: dict = {**BaseSTreeDSolver._parameter_constraints}

    def __init__(self, 
                 max_depth : int = 3,
                 max_num_nodes : Optional[int] = None,
                 min_leaf_node_size : int = 1,
                 time_limit : float = 600,
                 feature_ordering : str = "gini", 
                 hyper_tune: bool = False,
                 use_branch_caching: bool = False,
                 use_dataset_caching: bool = True,
                 use_terminal_solver: bool = True,
                 use_similarity_lower_bound: bool = True,
                 use_upper_bound: bool = True,
                 use_lower_bound: bool = True,
                 upper_bound: float = 2**31 - 1,
                 verbose : bool = False,
                 random_seed : int = 27,
                 continuous_binarize_strategy: str = 'quantile',
                 n_thresholds: int = 5,
                 n_categories: int = 5,
                 max_num_binary_features: int = None):
        """
        Construct a STreeDInstanceCostSensitiveClassifier

        Parameters:
            max_depth: the maximum depth of the tree
            max_num_nodes: the maximum number of branching nodes of the tree
            min_leaf_node_size: the minimum number of training instance that should end up in every leaf node
            time_limit: the time limit in seconds for fitting the tree
            feature_ordering: heuristic for the order that features are checked. Default: "gini", alternative: "in-order": the order in the given data
            hyper_tune: Use five-fold validation to tune the size of the tree to prevent overfitting
            use_branch_caching: Enable/Disable branch caching (typically the slower caching strategy. May be faster in some scenario's)
            use_dataset_caching: Enable/Disable dataset caching (typically the faster caching strategy)
            use_terminal_solver: Enable/Disable the depth-two solver (Enabled typically results in a large runtime advantage)
            use_similarity_lower_bound: Enable/Disable the similarity lower bound (Enabled typically results in a large runtime advantage)
            use_upper_bound: Enable/Disable the use of upper bounds (Enabled is typically faster)
            use_lower_bound: Enable/Disable the use of lower bounds (Enabled is typically faster)
            upper_bound: Search for a tree better than the provided upper bound
            verbose: Enable/Disable verbose output
            random_seed: the random seed used by the solver (for example when creating folds)
            continuous_binarization_strategy: the strategy used for binarizing continuous features
            n_thresholds: the number of thresholds to use per continuous feature
            n_categories: the number of categories to use per categorical feature
            max_num_binary_features: the maximum number of binary features (selected by random forest feature importance)
        """
        BaseSTreeDSolver.__init__(self, "instance-cost-sensitive", 
            max_depth=max_depth,
            max_num_nodes=max_num_nodes,
            min_leaf_node_size=min_leaf_node_size,
            time_limit = time_limit,
            cost_complexity=0,
            feature_ordering=feature_ordering,
            hyper_tune = hyper_tune,
            use_branch_caching=use_branch_caching,
            use_dataset_caching=use_dataset_caching,
            use_terminal_solver=use_terminal_solver,
            use_similarity_lower_bound=use_similarity_lower_bound,
            use_upper_bound=use_upper_bound,
            use_lower_bound=use_lower_bound,
            upper_bound=upper_bound,
            verbose=verbose,
            random_seed=random_seed,
            continuous_binarize_strategy=continuous_binarize_strategy,
            n_thresholds=n_thresholds,
            n_categories=n_categories,
            max_num_binary_features=max_num_binary_features)
        self.n_classes_ = None

    def _initialize_param_handler(self):
        super()._initialize_param_handler()
    
    def _check_data(self, y_in, reset=True):
        y_out = []
        if isinstance(y_in, pd.DataFrame):
            y_in = y_in.values
        if isinstance(y_in, np.ndarray):
            extra_data = []
            if len(y_in.shape) != 2 or y_in.shape[1] < 2:
                raise ValueError(f"The y-array does not have the right shape. Expected shape is a two-dimensional array with at least 2 columns.")
            if reset or self.n_classes_ is None:
                n_classes = y_in.shape[1]
            else:
                n_classes = self.n_classes_
                if n_classes != y_in.shape[1]:
                    raise ValueError(f"The y-array does not have the expected shape. Expected {n_classes} columns, but found {y_in.shape[1]}.")
            for row in y_in:
                extra_data.append(CostVector(row))
        else:
            extra_data = y_in
        
        num_labels = 0
        for i, inst in enumerate(extra_data):
            if not isinstance(inst, CostVector):
                raise ValueError("Each instance in y is expected to be of class CostVector.")
            y_out.append(np.argmin(inst.costs))
            num_labels = max(num_labels, len(inst.costs))

        if reset or self.n_classes_ is None:
            self.n_classes_ = num_labels
        return np.array(y_out, dtype=np.intc), extra_data

    def _process_extra_data(self, X, extra_data):
        return extra_data

    def fit(self, X, y, categorical=None):
        """
        Fits a STreeD model to the given training data.

        Args:
            x : array-like, shape = (n_samples, n_features)
            Data matrix

            y : array-like with shape (n_samples, n_labels) 
            Cost vector. For each instance specify the classification costs per label

            categorical : array-like, 
            List of column names that are categorical

        Returns:
            STreeDInstanceCostSensitiveClassifier

        Raises:
            ValueError: If x or y is None or if they have different number of rows.
        """
        y, extra_data = self._check_data(y)
        return super().fit(X, y, extra_data)
    
    def predict(self, X, extra_data=None):
        """
        Predicts the target variable for the given input feature data.

        Args:
            x : array-like, shape = (n_samples, n_features)
            Data matrix
            extra_data (optional) :array-like with shape (n_samples, n_labels) 
            Cost vector. For each instance specify the classification costs per label
            
        Returns:
            numpy.ndarray: A 1D array that represents the predicted target variable of the test data.
                The i-th element in this array corresponds to the predicted target variable for the i-th instance in `x`.
        """
        if extra_data is None:
            extra_data = [CostVector(np.zeros(self.n_classes_)) for i in range(len(X))]
        _, extra_data = self._check_data(extra_data, reset=False)
        return super().predict(X, extra_data)
    
    def score(self, X, y_test):
        """
        Computes the score for the given input feature data

        Args:
            x : array-like, shape = (n_samples, n_features)
            Data matrix
            y_test : array-like with shape (n_samples, n_labels) 
            Cost vector. For each instance specify the classification costs per label

        Returns:
            The score
        """
        y_test, extra_data = self._check_data(y_test, reset=False)
        return super().score(X, y_test, extra_data)

    def export_dot(self, filename, feature_names=None, label_names=None):
        """
        Export the tree as a .dot file (for plotting)
        """
        self.__export_directory = Path(filename).parent.absolute() / "leaf-distributions"
        return super().export_dot(filename, feature_names, label_names)
    
    def _get_branching_color(self, cost_sums):
        if not hasattr(self, "colors"):
            self.colors = _color_brew(self.n_classes_)
        best_ix = int(np.argmin(cost_sums))
        color = list(self.colors[best_ix])
        total_costs = sum(cost_sums)
        if cost_sums[best_ix] < 0.01 * total_costs:
            alpha = 0.0
        else:
            sorted_costs = sorted(cost_sums)
            if sorted_costs[1] < 0.01:
                alpha = 0.0
            else:
                alpha = (sorted_costs[1] - sorted_costs[0]) / (sorted_costs[1])
        color = [int(round(alpha * c + (1 - alpha) * 255, 0)) for c in color]
        # Return html color code in #RRGGBB format
        return "#%2x%2x%2x" % tuple(color)

    def _export_bar_plot(self, fh, node_id, label_names, train_data, text_prefix=""):
        if not hasattr(self, "_colors"):
            self._colors = _color_brew(self.n_classes_)
        import matplotlib.pyplot as plt
        plt.rc('axes', labelsize=6)
        plt.rc('xtick', labelsize=6)
        plt.rc('ytick', labelsize=6)
        plt.rc("axes", titlesize=7)
        cost_sums = np.zeros(self.n_classes_)
        costs = train_data[2]
        n = len(costs)
        for row in costs:
            for label in range(self.n_classes_):
                cost_sums[label] += row.costs[label]
        cost_sums /= n
        best_label = int(np.argmin(cost_sums))
        color = self._colors[best_label]

        colors = ['tab:gray' for label in range(self.n_classes_)]
        colors[best_label] = [c / 256 for c in color]
        _label_names = [self._get_label_str(i, label_names) for i in range(self.n_classes_)]
        label_costs = zip(_label_names, cost_sums, colors)
        label_costs = sorted(label_costs, key=lambda i: i[1])[:4]
        _label_names, cost_sums, colors = [list(i) for i in list(zip(*label_costs))]
        _label_names.reverse()
        cost_sums.reverse()
        colors.reverse()

        plt.figure(figsize=(1.8,1.2))
        plt.barh(_label_names, cost_sums, color=colors)
        label =  self._get_label_str(best_label, label_names)
        plt.title(f"{text_prefix} N = {n}, Best = {label}, Costs = {_dynamic_float_formatter(min(cost_sums))}")

        filename = self.__export_directory / f"leaf_ics_{node_id}.png"
        if not os.path.exists(self.__export_directory):
            os.makedirs(self.__export_directory)
        plt.savefig(filename, bbox_inches="tight", dpi=600)
        
        fh.write(f"{node_id}  [label=\"\", image=\"{filename}\", width=2.0, height=1.3, fixedsize=true, shape=none] ;\n")
    
    def _export_dot_predicate_node(self, fh, node, node_id, feature_names, label_names, train_data):
        predicate = self._get_predicate_str(node.feature, feature_names)
        try:
            self._export_bar_plot(fh, node_id, label_names, train_data, f"{predicate}\n")
        except:
            cost_sums = np.zeros(self.n_classes_)
            costs = train_data[2]
            n = len(costs)
            for row in costs:
                for label in range(self.n_classes_):
                    cost_sums[label] += row.costs[label]
            cost_sums /= n
            best_ix = int(np.argmin(cost_sums))
            best_label = self._get_label_str(best_ix, label_names)
            best_costs = cost_sums[best_ix]
            branching_color = self._get_branching_color(cost_sums)
            fh.write(f"{node_id} [label=\"{predicate}\nN = {n}, Best = {best_label}, Costs = {_dynamic_float_formatter(best_costs)}\", color=\"#222222\", fillcolor=\"{branching_color}\"] ;\n") 

    def _export_dot_leaf_node(self, fh, node, node_id, label_names, train_data):
        try:
            self._export_bar_plot(fh, node_id, label_names, train_data)
        except:
            if not hasattr(self, "_colors"):
                self._colors = _color_brew(self.n_classes_)
            color = self._colors[node.label]
            label =  self._get_label_str(node.label, label_names)
            hex_color = "#{:02x}{:02x}{:02x}".format(*color)
            hex_line_color = "#{:02x}{:02x}{:02x}".format(*[int(0.4 * c) for c in color])
            fh.write(f"{node_id}  [label=\"{label}\", color=\"{hex_line_color}\" fillcolor=\"{hex_color}\"] ;\n")